"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
from typing import Dict, List, Optional, Set, Union

from pydantic import ValidationError

from .models_step import Step, StepPushCommit


RETAG_ERROR_MESSAGE = "Multi-platform build steps cannot re-tag images. The following images are re-tagged:"
RUN_MP_ERROR_MESSAGE = "run is not allowed in the same step as a multi-platform build"
BUILD_MP_CACHE_ERROR_MESSAGE = "The cache_from arg must be a str, dict or list(dict) in the multi-platform build step"


class StepImagesInfo:
    def __init__(
        self, source_image: str, dest_images: List[str], is_multi_platform: bool
    ) -> None:
        self._is_multi_platform = is_multi_platform
        self._source_image = source_image
        self._dest_images = dest_images

    @property
    def is_multi_platform(self) -> bool:
        return self._is_multi_platform

    @property
    def source_image(self) -> str:
        return self._source_image

    @property
    def dest_images(self) -> list:
        return self._dest_images


def get_validation_errors(exc: ValidationError) -> List[str]:
    """Get validation errors to an Errors object"""
    errors = []
    for error in exc.errors():
        field = ".".join(str(item) for item in error["loc"])
        if error["type"] == "value_error.extra":
            errors.append(
                f"  {field}:  not a valid field, please check the spelling and documentation"
            )
        else:
            errors.append(f"  {field}:  {error['msg']} ({error['type']})")
    return errors


def validate_push(
    push: List[Union[str, StepPushCommit]],
    mp_push_tags: Set[str],
    step_name: str,
    update_mp_push_tags: bool = True,
) -> None:
    """
    Validate push step to ensure duplicate MP image tags are not used.

    Args:
        push (List[StepPushDict]): Push step
        mp_push_tags (Set[str]): Set of all tags used in multi-platform build steps
        step_name (str): Name of the step
        update_mp_push_tags (bool, optional): Whether to update the set of tags used in multi-platform steps.

    Raises:
        ValueError: If the config file is invalid
    """
    if not push:
        return
    for push_config in push:
        if not push_config.tags:
            tags = [push_config.repository]
        else:
            tags = [f"{push_config.repository}:{tag}" for tag in push_config.tags]
        for tag in tags:
            if tag in mp_push_tags:
                raise ValueError(
                    f"Cannot specify duplicate tag {tag} in build step {step_name}"
                )
            if update_mp_push_tags:
                mp_push_tags.add(tag)


def _get_base_image_from_dockerfile(dockerfile: str) -> str:
    """
    Get the base image from a dockerfile

    Args:
        dockerfile (str): Dockerfile

    Returns:
        str: Base image
    """
    dockerfile_text = None
    src_image = None
    if dockerfile:
        if os.path.exists(dockerfile):
            with open(dockerfile, "r") as file:
                dockerfile_text = file.read()
        else:
            # Dockerfile is defined as a string in the config file
            dockerfile_text = dockerfile

    if not isinstance(dockerfile_text, str):
        raise ValueError(f"There is an issue with the dockerfile {dockerfile}")

    for line in dockerfile_text.lower().strip().splitlines():
        if line.startswith("from"):
            image_name = line.replace("from", "").strip()
            if ":" not in image_name:
                image_name = f"{image_name}:latest"
            src_image = image_name

    if not src_image:
        raise ValueError(f"There is an issue with the dockerfile {dockerfile}")

    return src_image.lower()


def _get_source_image(step: Step) -> str:
    """
    Get the source image from the step.build or step.run.image

    Args:
        step (Step): Build step

    Returns:
        Optional[str]: source image
    """

    src_image = None

    # Gets the source image from build.dockerfile
    if step.build:
        dockerfile = step.build.dockerfile
        if not dockerfile and step.build.path:
            dockerfile = f"{step.build.path}/Dockerfile"
        src_image = _get_base_image_from_dockerfile(dockerfile)

    # Get the source image from step.run.image if build.dockerfile is not defined
    elif step.run and step.run.image:
        src_image = step.run.image
        if ":" not in src_image:
            src_image = f"{src_image}:latest"

    return src_image


def _get_destination_images(step: Step) -> List[str]:
    """
    Get the destination images from step.push or step.commit

    Args:
        step (Step): Build step

    Returns:
        List[str]: List of destination images
    """

    def _get_images(
        push_commit: List[StepPushCommit],
    ) -> List[str]:
        if not push_commit:
            return []
        current_images = []
        for item in push_commit:
            if item.tags:
                for tag in item.tags:
                    current_images.append(f"{item.repository}:{tag}")
            else:
                current_images.append(item.repository)
        return current_images

    images = _get_images(step.push)
    images.extend(_get_images(step.commit))
    return images


def validate_multiplatform_are_not_retagged(steps: Dict[str, Step]):
    """
    Validate multi-platform are not re-tagged

    Raises:
        ValueError: If the config file is invalid
    """
    step_images = {}

    # Iterate through each step and get information about the images
    for step_name, step in steps.items():
        source_image = _get_source_image(step)
        dest_images = _get_destination_images(step)
        step_images[step_name] = StepImagesInfo(
            source_image=source_image,
            dest_images=dest_images,
            is_multi_platform=step.is_multi_platform(),
        )

    # Iterate through each step images and check for multi-platform re-tagging
    retagged_images = []
    for step_name, step_images_info in step_images.items():
        if (
            # Ignore steps that do not push images
            not step_images_info.dest_images
            # Ignore steps that are also multi-platform, as re-tagging will work in this case
            or step_images_info.is_multi_platform
        ):
            continue

        other_steps_images_infos = [
            curr_step_images_info
            for curr_step_name, curr_step_images_info in step_images.items()
            if curr_step_name != step_name
        ]
        for other_step_info in other_steps_images_infos:
            if (
                step_images_info.source_image in other_step_info.dest_images
                and other_step_info.is_multi_platform
            ):
                retagged_images.append(step_images_info.source_image)

    if retagged_images:
        raise ValueError(f"{RETAG_ERROR_MESSAGE} {retagged_images}")


def validate_multiplatform_build(
    steps: Optional[Dict[str, Step]], mp_push_tags: Set[str]
):
    """
    Validate multi-platform build steps

    Args:
        steps (Optional[Dict[str]]): The steps to validate
        mp_push_tags (Set[str]): Set of all tags used in multi-platform build steps

    Raises:
        ValueError: If the config file is invalid
    """
    # Iterate through each step and validate multi-platform multi-platform steps
    for step_name, step in steps.items():
        if step.is_multi_platform():
            if step.build.platform is not None:
                raise ValueError(
                    f"Cannot specify both platform ({step.build.platform}) and "
                    f"platforms ({step.build.platforms}) in build step {step_name}"
                )

            if not isinstance(step.build.platforms, list):
                raise ValueError(f"platforms must be a list in build step {step_name}")

            if (
                step.build.platforms
                and isinstance(step.build.cache_from, list)
                and all(isinstance(x, str) for x in step.build.cache_from)
            ):
                raise ValueError(
                    f"{BUILD_MP_CACHE_ERROR_MESSAGE} {step_name} cannot be a list(str) for multiplatform images."
                )

            if step.build.import_param:
                raise ValueError(
                    f"import is not allowed in multi-platform build step {step_name}"
                )

            if step.run:
                raise ValueError(f"{RUN_MP_ERROR_MESSAGE} step {step_name}")

            # Check for valid push section, duplicate mp tags are not allowed
            validate_push(step.push, mp_push_tags, step_name)



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import json
import logging
import os
import re
import tarfile
import tempfile

import docker
import docker.errors

from buildrunner.errors import (
    BuildRunnerProcessingError,
)

from buildrunner.docker import get_dockerfile, new_client, force_remove_container

logger = logging.getLogger(__name__)

NO_CACHE_DEFAULT = False
RM_DEFAULT = True
PULL_DEFAULT = True


def build_image(
    path=None,
    inject=None,
    dockerfile=None,
    dockerd_url=None,
    timeout=None,
    docker_registry=None,
    temp_dir=None,
    console=None,
    nocache=NO_CACHE_DEFAULT,
    cache_from=None,
    rm=RM_DEFAULT,
    pull=PULL_DEFAULT,
    buildargs=None,
    platform=None,
    target=None,
    copy_source_path=True,
):
    """
    Build a Docker image using the DockerBuilder class.
    Args:
        path (str): Path to the build context.
        inject (dict): Dictionary of files to inject into the build context.
        dockerfile (str): Path to the Dockerfile to use.
        dockerd_url (str): URL to the Docker daemon.
        timeout (int): Timeout for the Docker client.
        docker_registry (str): Docker registry to use.
        temp_dir (str): Temporary directory to use.
        console (object): Console object to write output to.
        nocache (bool): Whether to use the Docker cache.
        cache_from (list): List of images to use as cache sources.
        rm (bool): Whether to remove intermediate containers.
        pull (bool): Whether to pull images from the registry.
        buildargs (dict): Build arguments to pass to the Docker client.
        platform (str): Platform to build the image for.
        target (str): Target stage to build.
        copy_source_path (bool): Whether to copy the source path into the build context.
    Returns:
        str: The ID of the built image.
    """
    logger.info("Using legacy builder")
    image = None
    builder = DockerBuilder(
        path,
        inject=inject,
        dockerfile=dockerfile,
        dockerd_url=dockerd_url,
        timeout=timeout,
        docker_registry=docker_registry,
        temp_dir=temp_dir,
        copy_source_path=copy_source_path,
    )
    try:
        exit_code = builder.build(
            console=console,
            nocache=nocache,
            cache_from=cache_from,
            rm=rm,
            pull=pull,
            buildargs=buildargs,
            platform=platform,
            target=target,
        )
        if exit_code != 0 or not builder.image:
            raise BuildRunnerProcessingError("Error building image")
        image = builder.image
    finally:
        builder.cleanup()
    return image


class DockerBuilder:  # pylint: disable=too-many-instance-attributes
    """
    An object that manages and orchestrates building a Docker image from
    a Dockerfile.
    """

    def __init__(
        self,
        path=None,
        inject=None,
        dockerfile=None,
        dockerd_url=None,
        timeout=None,
        docker_registry=None,
        temp_dir=None,
        copy_source_path=True,
    ):  # pylint: disable=too-many-arguments
        self.path = path
        self.inject = inject
        self.temp_dir = temp_dir
        self.dockerfile = None
        self.cleanup_dockerfile = False
        self.copy_source_path = copy_source_path

        self.dockerfile, self.cleanup_dockerfile = get_dockerfile(
            dockerfile, self.temp_dir
        )

        self.docker_client = new_client(
            dockerd_url=dockerd_url,
            timeout=timeout,
        )
        self.docker_registry = docker_registry
        self.image = None
        self.intermediate_containers = []

    @staticmethod
    def _sanitize_buildargs(buildargs=None):
        """
        Ensure that buildargs are correct for the Docker API.

        Args:

            :buildargs: (dict): unsanitized arguments to be passed to the Docker client.

        Returns:

            :dict: sanitized arguments to be passed to the Docker client.
        """
        if not isinstance(buildargs, dict):
            raise TypeError("buildargs must be a dictionary of keys/values")

        return {k: str(v) for k, v in list(buildargs.items())}

    # pylint: disable=too-many-branches,too-many-locals,too-many-arguments,invalid-name
    def build(
        self,
        console=None,
        nocache=NO_CACHE_DEFAULT,
        cache_from=None,
        rm=RM_DEFAULT,
        pull=PULL_DEFAULT,
        buildargs=None,
        platform=None,
        target=None,
    ):
        """
        Run a docker build using the configured context, constructing the
        context tar file if necessary.
        """
        logger.info("Using legacy builder")
        if cache_from is None:
            cache_from = []
        if buildargs is None:
            buildargs = {}

        # Always add default registry to build args
        buildargs["DOCKER_REGISTRY"] = self.docker_registry
        stream = None

        # create our own tar file, injecting the appropriate paths
        # pylint: disable=consider-using-with
        if self.inject or not self.path:
            logger.info(
                "[Warning] When injecting files into the build context the .dockerignore is not used."
            )

            _fileobj = tempfile.NamedTemporaryFile(dir=self.temp_dir)
            with tarfile.open(mode="w", fileobj=_fileobj) as tfile:
                if self.path and self.copy_source_path:
                    tfile.add(self.path, arcname=".")
                if self.inject:
                    for to_inject, dest in self.inject.items():
                        tfile.add(to_inject, arcname=dest)
                if self.dockerfile:
                    tfile.add(self.dockerfile, arcname="./Dockerfile")
            _fileobj.seek(0)

            stream = self.docker_client.build(
                path=None,
                nocache=nocache,
                cache_from=cache_from,
                custom_context=True,
                fileobj=_fileobj,
                rm=rm,
                pull=pull,
                buildargs=self._sanitize_buildargs(buildargs),
                platform=platform,
                target=target,
            )
        else:
            stream = self.docker_client.build(
                path=self.path,
                nocache=nocache,
                cache_from=cache_from,
                rm=rm,
                pull=pull,
                buildargs=self._sanitize_buildargs(buildargs),
                platform=platform,
                target=target,
                dockerfile=self.dockerfile,
            )

        # monitor output for logs and status
        exit_code = 0
        msg_buffer = ""
        for msg_str in stream:  # pylint: disable=too-many-nested-blocks
            for msg in msg_str.decode("utf-8").split("\n"):
                if msg:
                    msg_buffer += msg
                    try:
                        # there is a limit on the chars returned in the stream
                        # generator, so if we don't have a valid json message
                        # here we get the next msg and append to the current
                        # one
                        json_msg = json.loads(msg_buffer)
                        msg_buffer = ""
                    except ValueError:
                        continue
                    if "stream" in json_msg:
                        # capture intermediate containers for cleanup later
                        # the command line 'docker build' has a '--force-rm' option,
                        # but that isn't available in the python client
                        container_match = re.search(
                            r" ---> Running in ([0-9a-f]+)",
                            json_msg["stream"],
                        )
                        if container_match:
                            self.intermediate_containers.append(
                                container_match.group(1)
                            )

                        # capture the resulting image
                        image_match = re.search(
                            r"Successfully built ([0-9a-f]+)",
                            json_msg["stream"],
                        )
                        if image_match:
                            self.image = image_match.group(1)

                        if console:
                            console.write(json_msg["stream"])
                    if "error" in json_msg:
                        exit_code = 1
                        if "errorDetail" in json_msg:
                            if "message" in json_msg["errorDetail"] and console:
                                console.write(f"{json_msg['errorDetail']['message']}\n")

        return exit_code

    def cleanup(self):
        """
        Cleanup the docker build environment.
        """
        # cleanup the generated dockerfile if present
        if self.cleanup_dockerfile:
            if self.dockerfile and os.path.exists(self.dockerfile):
                os.remove(self.dockerfile)

        # iterate through and destroy intermediate containers
        for container in self.intermediate_containers:
            try:
                force_remove_container(self.docker_client, container)
            except docker.errors.APIError as err:
                logger.debug(
                    f"Error removing intermediate container {container}: {err}"
                )



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
import tempfile
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field, field_validator, ValidationError

from .models_step import Step, StepPushSecurityScanConfig
from .validation import (
    get_validation_errors,
    validate_multiplatform_build,
    validate_multiplatform_are_not_retagged,
    validate_push,
)


DEFAULT_CACHES_ROOT = "~/.buildrunner/caches"
# Marker for using the local registry instead of an upstream registry
MP_LOCAL_REGISTRY = "local"
DEFAULT_TO_LEGACY_BUILDER = True


class GithubModel(BaseModel, extra="forbid"):
    endpoint: str
    version: str = None
    username: str = os.getenv("USER", os.getenv("LOGNAME"))
    app_token: str = ""


class SSHKey(BaseModel, extra="forbid"):
    file: Optional[str] = None
    key: Optional[str] = None
    password: Optional[str] = None
    prompt_password: Optional[bool] = Field(alias="prompt-password", default=None)
    aliases: Optional[List[str]] = None


class DockerBuildCacheConfig(BaseModel, extra="forbid"):
    builders: Optional[List[str]] = None
    from_config: Optional[Union[dict, str]] = Field(None, alias="from")
    to_config: Optional[Union[dict, str]] = Field(None, alias="to")


class GlobalSecurityScanConfig(BaseModel, extra="forbid"):
    """
    The global config version of security scan configuration has default values.
    """

    enabled: bool = False
    scanner: str = "trivy"
    version: str = "latest"
    # The local cache directory for the scanner (used if supported by the scanner)
    cache_dir: Optional[str] = Field(None, alias="cache-dir")
    config: dict = {
        "timeout": "20m",
        # Do not error on vulnerabilities by default
        "exit-code": 0,
    }
    max_score_threshold: Optional[float] = Field(None, alias="max-score-threshold")

    def merge_scan_config(
        self, push_config: Optional[StepPushSecurityScanConfig]
    ) -> "GlobalSecurityScanConfig":
        """
        Merges the push security scanning config with the global security config
        and returns a merged config without modifying the global configuration.
        :param push_config: the step push security scan config
        :return: a merged scan config
        """
        if not push_config:
            return self
        new_model = {
            field: getattr(push_config, field)
            for field, value in self.model_fields.items()
            if getattr(push_config, field, None) is not None
        }
        # model_copy does not handle actually copying the config dict, so copy this manually and update with
        # the step config once the model copy is done
        config_value = new_model.pop("config", None)
        new_model["config"] = deepcopy(self.config)
        cloned_config = self.model_copy(update=new_model)
        if config_value:
            cloned_config.config.update(config_value)
        return cloned_config


class GlobalConfig(BaseModel, extra="forbid"):
    """Top level global config model"""

    github: Optional[Dict[str, GithubModel]] = None
    env: Optional[Dict[str, Any]] = None
    build_servers: Optional[Dict[str, List[str]]] = Field(
        alias="build-servers", default=None
    )
    ssh_keys: Optional[List[SSHKey]] = Field(alias="ssh-keys", default=None)
    local_files: Optional[Dict[str, str]] = Field(alias="local-files", default=None)
    docker_build_cache: Optional[DockerBuildCacheConfig] = Field(
        alias="docker-build-cache", default=DockerBuildCacheConfig()
    )
    caches_root: Optional[str] = Field(alias="caches-root", default=DEFAULT_CACHES_ROOT)
    # Default to docker.io if none is configured
    docker_registry: Optional[str] = Field(alias="docker-registry", default="docker.io")
    """
    Get temp dir in the following priorities:
    * Environment variable
    * Global configuration property
    * Configured system temp directory
    """
    temp_dir: Optional[str] = Field(
        alias="temp-dir",
        default=os.getenv("BUILDRUNNER_TEMPDIR", tempfile.gettempdir()),
    )
    disable_multi_platform: Optional[bool] = Field(
        alias="disable-multi-platform", default=None
    )
    build_registry: Optional[str] = Field(
        alias="build-registry", default=MP_LOCAL_REGISTRY
    )
    platform_builders: Optional[Union[Dict[str, str], Dict[str, List[str]]]] = Field(
        alias="platform-builders", default=None
    )
    security_scan: GlobalSecurityScanConfig = Field(
        GlobalSecurityScanConfig(), alias="security-scan"
    )

    @field_validator("ssh_keys", mode="before")
    @classmethod
    def transform_ssh_keys(cls, val) -> Optional[List[dict]]:
        if not isinstance(val, dict):
            return val
        return [val]


class Config(BaseModel, extra="forbid"):
    """Top level config model"""

    version: Optional[float] = None
    use_legacy_builder: Optional[bool] = Field(
        alias="use-legacy-builder", default=DEFAULT_TO_LEGACY_BUILDER
    )
    steps: Dict[str, Step]

    @field_validator("steps")
    @classmethod
    def validate_steps(cls, vals, info) -> None:
        """
        Validate the config file

        Raises:
            ValueError : If the config file is invalid
        """

        if not vals:
            raise ValueError('The "steps" configuration was not provided')

        # Checks steps for mutli-platform or secrets
        has_multi_platform_build = False

        #  Check for multi-platform builds and secrets validation
        for step in vals.values():
            has_multi_platform_build = (
                has_multi_platform_build or step.is_multi_platform()
            )

            # If the step has secrets and the builder is legacy or no platforms are set for the step, raise an error
            if (
                step.has_secrets()
                and not step.is_multi_platform()
                and info.data.get("use_legacy_builder")
            ):
                raise ValueError(
                    "Build secrets are not supported with the legacy builder. Please set use-legacy-builder to false"
                    " or add platforms to the build section in order to use secrets in your build."
                )

        if has_multi_platform_build:
            mp_push_tags = set()
            validate_multiplatform_build(vals, mp_push_tags)
            validate_multiplatform_are_not_retagged(vals)

            # Validate that all tags are unique across all multi-platform step
            for step_name, step in vals.items():
                # Check that there are no single platform tags that match multi-platform tags
                if not step.is_multi_platform():
                    if step.push is not None:
                        validate_push(
                            push=step.push,
                            mp_push_tags=mp_push_tags,
                            step_name=step_name,
                            update_mp_push_tags=False,
                        )
        return vals


def generate_and_validate_config(
    **kwargs,
) -> Tuple[Optional[Config], Optional[List[str]]]:
    try:
        return Config(**kwargs), None
    except ValidationError as exc:
        return None, get_validation_errors(exc)


def generate_and_validate_global_config(
    **kwargs,
) -> Tuple[Optional[GlobalConfig], Optional[List[str]]]:
    try:
        return GlobalConfig(**kwargs), None
    except ValidationError as exc:
        return None, get_validation_errors(exc)



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""
# pylint: disable=too-many-lines

from collections import OrderedDict
import fnmatch
import importlib.metadata
import inspect
import json
import logging
import os
import shutil
import sys
import tarfile
import tempfile
import traceback
from typing import List, Optional

import requests

from retry import retry
from vcsinfo import detect_vcs, VCSUnsupported, VCSMissingRevision
from docker.errors import ImageNotFound

from buildrunner import docker, loggers
from buildrunner.config import (
    BuildRunnerConfig,
)
from buildrunner.config.models import DEFAULT_CACHES_ROOT
from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProcessingError,
    BuildRunnerError,
)
from buildrunner.steprunner import BuildStepRunner
from buildrunner.docker.multiplatform_image_builder import MultiplatformImageBuilder
import buildrunner.docker.builder


LOGGER = logging.getLogger(__name__)

try:
    __version__ = importlib.metadata.version("buildrunner")
except importlib.metadata.PackageNotFoundError:
    __version__ = "DEVELOPMENT"

SOURCE_DOCKERFILE = os.path.join(os.path.dirname(__file__), "SourceDockerfile")


class BuildRunner:
    """
    Class used to manage running a build.
    """

    def __init__(
        self,
        *,
        build_dir: str,
        build_results_dir: str,
        global_config_file: Optional[str],
        run_config_file: Optional[str],
        build_time: int,
        build_number: int,
        push: bool,
        cleanup_images: bool,
        cleanup_cache: bool,
        steps_to_run: Optional[List[str]],
        publish_ports: bool,
        log_generated_files: bool,
        docker_timeout: int,
        local_images: bool,
        platform: Optional[str],
        global_config_overrides: dict,
        container_labels: Optional[str] = None,
    ):  # pylint: disable=too-many-statements,too-many-branches,too-many-locals,too-many-arguments
        self.build_dir = build_dir
        self.build_results_dir = build_results_dir
        self.build_time = build_time
        self.build_number = build_number
        self.push = push
        self.cleanup_images = cleanup_images
        self.cleanup_cache = cleanup_cache
        self.generated_images = []
        self.removed_images = set()
        # The set of images (including tag) that were committed as part of this build
        # This is used to check if images should be pulled by default or not
        self.committed_images = set()
        self.repo_tags_to_push = []
        self.steps_to_run = steps_to_run
        self.publish_ports = publish_ports
        self.log_generated_files = log_generated_files
        self.docker_timeout = docker_timeout
        self.local_images = local_images
        self.platform = platform

        self.tmp_files = []
        self.artifacts = OrderedDict()
        self.pypi_packages = OrderedDict()

        self.exit_code = None
        self._source_image = None
        self._source_archive = None
        self._log = None
        self._step_runner = None

        try:
            vcs = detect_vcs(self.build_dir)
            self.build_id = f"{vcs.id_string}-{self.build_number}"
        except VCSUnsupported as err:
            self.log.write(
                f"{err}\nPlease verify you have a VCS set up for this project.\n"
            )
            sys.exit()
        except VCSMissingRevision as err:
            self.log.write(f"{err}\nMake sure you have at least one commit.\n")
            sys.exit()

        # load global configuration - must come *after* VCS detection
        BuildRunnerConfig.initialize_instance(
            push=push,
            build_number=self.build_number,
            build_id=self.build_id,
            vcs=vcs,
            steps_to_run=self.steps_to_run,
            build_dir=self.build_dir,
            global_config_file=global_config_file,
            run_config_file=run_config_file,
            log_generated_files=self.log_generated_files,
            build_time=self.build_time,
            tmp_files=self.tmp_files,
            global_config_overrides=global_config_overrides,
            container_labels=container_labels,
            platform=platform,
        )
        self.buildrunner_config = BuildRunnerConfig.get_instance()

        # cleanup local cache
        if self.cleanup_cache:
            self.clean_cache()

        if steps_to_run:
            missing_steps = [
                step
                for step in steps_to_run
                if step not in self.buildrunner_config.run_config.steps
            ]
            if missing_steps:
                raise BuildRunnerConfigurationError(
                    f"The following steps do not exist: {', '.join(missing_steps)}"
                )

    @property
    def log(self) -> loggers.ConsoleLogger:
        """
        Create the log file and open for writing
        """
        if self._log is None:
            self._log = loggers.ConsoleLogger(__name__)
            self.add_artifact(
                os.path.basename(
                    loggers.get_build_log_file_path(self.build_results_dir)
                ),
                {"type": "log"},
            )
        return self._log

    @staticmethod
    def get_cache_archive_ext():
        """
        Returns the archive file extension used for cache archive files
        """
        return "tar"

    def get_cache_archive_file(self, cache_name, project_name=""):
        """
        Given a cache name determine the local file path.
        """

        def get_filename(caches_root, cache_name):
            local_cache_archive_file = os.path.expanduser(
                os.path.join(caches_root, cache_name)
            )
            cache_dir = os.path.dirname(local_cache_archive_file)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)
            return local_cache_archive_file

        cache_name = f"{cache_name}.{self.get_cache_archive_ext()}"
        if project_name != "":
            cache_name = f"{project_name}-{cache_name}"

        caches_root = BuildRunnerConfig.get_instance().global_config.caches_root
        try:
            local_cache_archive_file = get_filename(caches_root, cache_name)
        except Exception as exc:  # pylint: disable=broad-except
            # Intentionally catch all exceptions here since we don't want to fail the build
            LOGGER.warning(f"There was an issue with {caches_root}: {str(exc)}")
            local_cache_archive_file = get_filename(DEFAULT_CACHES_ROOT, cache_name)
            LOGGER.warning(f"Using {DEFAULT_CACHES_ROOT} for the cache directory")

        return local_cache_archive_file

    @staticmethod
    def clean_cache():
        """
        Clean cache dir
        """
        global_config = BuildRunnerConfig.get_instance().global_config
        cache_dir = os.path.expanduser(global_config.caches_root)
        if os.path.exists(cache_dir):
            LOGGER.info(f'Cleaning cache dir "{cache_dir}"')
            shutil.rmtree(f"{cache_dir}/")
            LOGGER.info(f'Cleaned cache dir "{cache_dir}"')
        else:
            LOGGER.info(f'Cache dir "{cache_dir}" is already clean')

    def add_artifact(self, artifact_file, properties):
        """
        Register a build artifact to be included in the artifacts manifest.
        """
        self.artifacts[artifact_file] = properties

    @retry(exceptions=FileNotFoundError, tries=5, delay=1, backoff=3, max_delay=10)
    def _create_archive_tarfile(self, dir_to_add, _fileobj, filter_func):
        """
        Create the tarfile with retries
        """
        with tarfile.open(mode="w", fileobj=_fileobj) as tfile:
            tfile.add(dir_to_add, arcname="", filter=filter_func)

    def get_source_archive_path(self):
        """
        Create the source archive for use in remote builds or to build the
        source image.
        """
        if not self._source_archive:
            buildignore = os.path.join(self.build_dir, ".buildignore")
            excludes = []
            if os.path.exists(buildignore):
                with open(buildignore, "r", encoding="utf-8") as _file:
                    excludes = _file.read().splitlines()

            def _filter_results_and_excludes(tarinfo):
                """
                Filter to exclude results dir and listed excludes from source archive.
                """
                if tarinfo.name == os.path.basename(self.build_results_dir):
                    return None
                for _ex in excludes:
                    if _ex and _ex.strip() and fnmatch.fnmatch(tarinfo.name, _ex):
                        return None
                return tarinfo

            self.log.write("Creating source archive\n")
            _fileobj = None
            try:
                # pylint: disable=consider-using-with
                _fileobj = tempfile.NamedTemporaryFile(
                    delete=False,
                    dir=BuildRunnerConfig.get_instance().global_config.temp_dir,
                )
                self._create_archive_tarfile(
                    self.build_dir, _fileobj, _filter_results_and_excludes
                )
                self._source_archive = _fileobj.name
            finally:
                if _fileobj:
                    _fileobj.close()
        return self._source_archive

    def get_source_image(self):
        """
        Get and/or create the base image source containers will be created from.
        """
        if not self._source_image:
            self.log.write("Creating source image\n")
            source_archive_path = self.get_source_archive_path()
            inject = {
                source_archive_path: "source.tar",
                SOURCE_DOCKERFILE: "Dockerfile",
            }
            image = buildrunner.docker.builder.build_image(
                temp_dir=self.buildrunner_config.global_config.temp_dir,
                inject=inject,
                timeout=self.docker_timeout,
                docker_registry=self.buildrunner_config.global_config.docker_registry,
                nocache=True,
                pull=False,
            )
            self._source_image = image

        return self._source_image

    def _write_artifact_manifest(self):
        """
        If we have registered artifacts write the files and associated metadata
        to the artifacts manifest.
        """
        if self.artifacts:
            self.log.write("\nWriting artifact properties\n")
            artifact_manifest = os.path.join(
                self.build_results_dir,
                "artifacts.json",
            )
            # preserve contents of artifacts.json between steps run separately
            if os.path.exists(artifact_manifest):
                with open(artifact_manifest, "r", encoding="utf-8") as _af:
                    data = json.load(_af, object_pairs_hook=OrderedDict)
                    artifacts = OrderedDict(
                        list(data.items()) + list(self.artifacts.items())
                    )
            else:
                artifacts = self.artifacts

            with open(artifact_manifest, "w", encoding="utf-8") as _af:
                json.dump(artifacts, _af, indent=2)

    def _exit_message(self, exit_explanation):
        """
        Determine the exit message and output to the log.
        """
        if self.exit_code:
            exit_message = "Build ERROR."
            log_method = self.log.error
        else:
            exit_message = "Build SUCCESS."
            log_method = self.log.info

        if self.log:
            if exit_explanation:
                self.log.info("")
                log_method(exit_explanation)
            self.log.info("")
            log_method(exit_message)
        else:
            if exit_explanation:
                print(f"\n{exit_explanation}")
            print(f"\n{exit_message}")

    def run(self):  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
        """
        Run the build.
        """
        # reset the exit_code
        self.exit_code = None

        exit_explanation = None
        try:  # pylint: disable=too-many-nested-blocks
            with MultiplatformImageBuilder(
                docker_registry=self.buildrunner_config.global_config.docker_registry,
                build_registry=self.buildrunner_config.global_config.build_registry,
                temp_dir=self.buildrunner_config.global_config.temp_dir,
                platform_builders=self.buildrunner_config.global_config.platform_builders,
                cache_builders=self.buildrunner_config.global_config.docker_build_cache.builders,
                cache_from=self.buildrunner_config.global_config.docker_build_cache.from_config,
                cache_to=self.buildrunner_config.global_config.docker_build_cache.to_config,
            ) as multi_platform:
                self.get_source_archive_path()
                # run each step
                for (
                    step_name,
                    step_config,
                ) in self.buildrunner_config.run_config.steps.items():
                    # Reset the cache_from and cache_to for each step to the global values
                    multi_platform.set_cache_from(
                        self.buildrunner_config.global_config.docker_build_cache.from_config
                    )
                    multi_platform.set_cache_to(
                        self.buildrunner_config.global_config.docker_build_cache.to_config
                    )

                    if not self.steps_to_run or step_name in self.steps_to_run:
                        image_config = BuildStepRunner.ImageConfig(
                            self.local_images, self.platform
                        )

                        # Override the multiplatform cache_from and cache_to if the step has its own
                        if step_config.build and step_config.build.cache_from:
                            LOGGER.info(
                                f"Overriding cache_from with {step_config.build.cache_from}"
                            )
                            multi_platform.set_cache_from(step_config.build.cache_from)
                        if step_config.build and step_config.build.cache_to:
                            LOGGER.info(
                                f"Overriding cache_to with {step_config.build.cache_to}"
                            )
                            multi_platform.set_cache_to(step_config.build.cache_to)

                        self._step_runner = BuildStepRunner(
                            self,
                            step_name,
                            step_config,
                            image_config,
                            multi_platform,
                            self.buildrunner_config.container_labels,
                        )
                        self._step_runner.run()

                self.log.write(
                    "\nFinalizing build\n________________________________________\n"
                )

                # see if we should push registered tags to remote registries/repositories
                if self.push:
                    self.log.write(
                        "Push requested--pushing generated images/packages to remote registries/repositories\n"
                    )
                    # Push multi-platform images
                    if multi_platform.num_built_images:
                        self.log.write(
                            f"===> Pushing {multi_platform.num_built_images} multiplatform image(s)\n"
                        )
                        multi_platform.push()

                    # Push single platform images
                    _docker_client = docker.new_client(timeout=self.docker_timeout)
                    for _repo_tag, _insecure_registry in self.repo_tags_to_push:
                        self.log.write(f"\nPushing {_repo_tag}\n")

                        # Newer Python Docker bindings drop support for the insecure_registry
                        # option.  This test will optionally use it when it's available.
                        push_kwargs = {
                            "stream": True,
                        }
                        if (
                            "insecure_registry"
                            in inspect.getfullargspec(_docker_client.push).args
                        ):
                            push_kwargs["insecure_registry"] = _insecure_registry

                        stream = _docker_client.push(_repo_tag, **push_kwargs)
                        previous_status = None
                        for msg_str in stream:
                            for msg in msg_str.decode("utf-8").split("\n"):
                                if not msg:
                                    continue
                                msg = json.loads(msg)
                                if "status" in msg:
                                    if msg["status"] == previous_status:
                                        continue
                                    self.log.write(msg["status"] + "\n")
                                    previous_status = msg["status"]
                                elif "errorDetail" in msg:
                                    error_detail = (
                                        f"Error pushing image: {msg['errorDetail']}\n"
                                    )
                                    self.log.write("\n" + error_detail)
                                    self.log.write(
                                        (
                                            "This could be because you are not "
                                            "authenticated with the given Docker "
                                            "registry (try 'docker login "
                                            "<registry>')\n\n"
                                        )
                                    )
                                    raise BuildRunnerProcessingError(error_detail)
                                else:
                                    self.log.write(str(msg) + "\n")

                    # Push to pypi repositories
                    # Placing the import here avoids the dependency when pypi is not needed
                    import twine.commands.upload  # pylint: disable=import-outside-toplevel

                    for _, _items in self.pypi_packages.items():
                        twine.commands.upload.upload(
                            _items["upload_settings"], _items["packages"]
                        )
                else:
                    self.log.write("\nPush not requested\n")

        except requests.exceptions.ConnectionError as rce:
            print(str(rce))
            exit_explanation = (
                "Error communicating with the remote host.\n\tCheck that the "
                "remote docker Daemon is running and/or that the DOCKER_* "
                "environment variables are set correctly.\n\tCheck that the "
                "remote PyPi server information is set correctly."
            )
            self.exit_code = 1
        except ImageNotFound as inf:
            exit_explanation = f"Image not found: {inf.explanation}"
            self.exit_code = os.EX_CONFIG
        except BuildRunnerError as exc:
            exit_explanation = str(exc)
            self.exit_code = (
                os.EX_CONFIG if isinstance(exc, BuildRunnerConfigurationError) else 1
            )
        except Exception as exc:
            exit_explanation = f"Encountered unhandled exception ({type(exc).__name__}) {traceback.format_exc()}"
            self.exit_code = 1

        finally:
            self._write_artifact_manifest()

            _docker_client = docker.new_client(timeout=self.docker_timeout)

            # cleanup the source image
            if self._source_image:
                self.log.write(f"Destroying source image {self._source_image}\n")
                try:
                    _docker_client.remove_image(
                        self._source_image,
                        noprune=False,
                        force=True,
                    )
                except ImageNotFound:
                    self.log.warning(
                        f"Failed to remove source image {self._source_image}\n"
                    )

            if self.cleanup_images:
                self.log.write("Removing local copy of generated images\n")
                # cleanup all registered docker images
                # reverse the order of the images since child images would likely come after parent images
                for _image in self.generated_images[::-1]:
                    try:
                        if _image in self.removed_images:
                            self.log.write(f"Image {_image} already removed\n")
                            continue
                        _docker_client.remove_image(
                            _image,
                            noprune=False,
                            force=True,
                        )
                        self.removed_images.add(_image)
                    except Exception as _ex:  # pylint: disable=broad-except
                        self.log.write(f"Unable to remove image {_image}: {str(_ex)}\n")
            else:
                self.log.write("Keeping generated images\n")
            if self._source_archive:
                self.log.write("Destroying source archive\n")
                os.remove(self._source_archive)

            # remove any temporary files that we created
            for tmp_file in self.tmp_files:
                if os.path.exists(tmp_file):
                    os.remove(tmp_file)

            self._exit_message(exit_explanation)


# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import logging
import os
import tempfile
import time
from typing import Dict, List, Optional

import yaml

import buildrunner.docker
from buildrunner.config import BuildRunnerConfig
from buildrunner.config.models import GlobalSecurityScanConfig
from buildrunner.config.models_step import StepPushCommit, StepPushSecurityScanConfig
from buildrunner.docker.image_info import BuiltImageInfo
from buildrunner.docker.multiplatform_image_builder import MultiplatformImageBuilder
from buildrunner.docker.runner import DockerRunner
from buildrunner.errors import (
    BuildRunnerProcessingError,
)
from buildrunner.steprunner.tasks import BuildStepRunnerTask
from buildrunner.utils import sanitize_tag


LOGGER = logging.getLogger(__name__)
ARTIFACT_SECURITY_SCAN_KEY = "docker:security-scan"


class RepoDefinition:
    """
    Contains the definition for a push repository.
    """

    def __init__(
        self,
        repository: str,
        tags: Optional[List[str]],
        security_scan: Optional[StepPushSecurityScanConfig],
    ):
        self.security_scan = security_scan
        # Force a lower-case repo
        repo_lower = repository.lower()
        if repo_lower != repository:
            LOGGER.info(
                f"Forcing repository to lowercase: {repository} => {repo_lower}"
            )
        self.repository = repo_lower

        if tags:
            self.tags = [sanitize_tag(tag) for tag in tags]
        else:
            self.tags = []

        # if there is a tag in the repository value, strip it off and add the tag to the list of tags
        tag_index = self.repository.find(":")
        if tag_index > 0:
            tag = self.repository[tag_index + 1 :]
            if tag not in self.tags:
                self.tags.append(tag)
            self.repository = self.repository[0:tag_index]


class PushBuildStepRunnerTask(BuildStepRunnerTask):
    """
    Class used to push the resulting image (either from the build task, or if
    there is a run task, the snapshot of the resulting run container) to the
    given registry/repository.
    """

    def __init__(self, step_runner, pushes: List[StepPushCommit], commit_only=False):
        super().__init__(step_runner, pushes[0])
        self._docker_client = buildrunner.docker.new_client(
            timeout=step_runner.build_runner.docker_timeout,
        )
        self._commit_only = commit_only
        self._repos = [
            RepoDefinition(
                push.repository,
                push.tags,
                push.security_scan,
            )
            for push in pushes
        ]

    def _security_scan_mp(
        self,
        built_image: BuiltImageInfo,
        log_image_ref: str,
        push_security_scan: Optional[StepPushSecurityScanConfig],
    ) -> Dict[str, dict]:
        """
        Does a security scan for each built image in a multiplatform image.
        :param built_image: the multiplatform built image info
        :param log_image_ref: the image label to log
        :param push_security_scan: the optional security scan overrides
        :return: a dictionary of platform names to security scan information (for each built platform image)
        """
        scan_results = {}
        for image in built_image.built_images:
            result = self._security_scan(
                repository=image.repo,
                tag=image.tag,
                log_image_ref=f"{log_image_ref}:{image.platform}",
                pull=True,
                push_security_scan=push_security_scan,
            )
            if result:
                scan_results[image.platform] = result
        if not scan_results:
            return {}
        return {ARTIFACT_SECURITY_SCAN_KEY: scan_results}

    def _security_scan_single(
        self,
        repo: str,
        tag: str,
        push_security_scan: Optional[StepPushSecurityScanConfig],
    ) -> Dict[str, dict]:
        log_image_ref = f"{repo}:{tag}"
        result = self._security_scan(
            repository=repo,
            tag=tag,
            log_image_ref=log_image_ref,
            pull=False,
            push_security_scan=push_security_scan,
        )
        if not result:
            return {}
        return {
            ARTIFACT_SECURITY_SCAN_KEY: {
                MultiplatformImageBuilder.get_native_platform(): result
            }
        }

    def _security_scan(
        self,
        *,
        repository: str,
        tag: str,
        log_image_ref: str,
        pull: bool,
        push_security_scan: Optional[StepPushSecurityScanConfig],
    ) -> Optional[dict]:
        # If the security scan is not enabled, do nothing
        security_scan_config = BuildRunnerConfig.get_instance().global_config.security_scan.merge_scan_config(
            push_security_scan
        )
        if log_image_ref != f"{repository}:{tag}":
            log_image_ref = f"{log_image_ref} ({repository}:{tag})"
        if not security_scan_config.enabled:
            LOGGER.debug(
                f"Image scanning is disabled, skipping scan of {log_image_ref}"
            )
            return None

        LOGGER.info(
            f"Scanning {log_image_ref} for security issues using {security_scan_config.scanner}"
        )

        if security_scan_config.scanner == "trivy":
            return self._security_scan_trivy(
                security_scan_config=security_scan_config,
                repository=repository,
                tag=tag,
                log_image_ref=log_image_ref,
                pull=pull,
            )
        raise Exception(f"Unsupported scanner {security_scan_config.scanner}")

    @staticmethod
    def _security_scan_trivy_parse_results(
        security_scan_config: GlobalSecurityScanConfig, results: dict
    ) -> dict:
        max_score = 0
        vulnerabilities = []
        for result in results.get("Results", []):
            if not result.get("Vulnerabilities"):
                continue
            for cur_vuln in result.get("Vulnerabilities"):
                score = cur_vuln.get("CVSS", {}).get("nvd", {}).get("V3Score")
                vulnerabilities.append({
                    "cvss_v3_score": score,
                    "severity": cur_vuln.get("Severity"),
                    "vulnerability_id": cur_vuln.get("VulnerabilityID"),
                    "pkg_name": cur_vuln.get("PkgName"),
                    "installed_version": cur_vuln.get("InstalledVersion"),
                    "primary_url": cur_vuln.get("PrimaryURL"),
                })
                if score:
                    max_score = max(max_score, score)

        if security_scan_config.max_score_threshold:
            if max_score >= security_scan_config.max_score_threshold:
                raise BuildRunnerProcessingError(
                    f"Max vulnerability score ({max_score}) is above the "
                    f"configured threshold ({security_scan_config.max_score_threshold})"
                )
            LOGGER.info(
                f"Max vulnerability score ({max_score}) is less than the "
                f"configured threshold ({security_scan_config.max_score_threshold})"
            )
        else:
            LOGGER.debug(
                f"Max vulnerability score is {max_score}, but no max score threshold is configured"
            )
        return {
            "max_score": max_score,
            "vulnerabilities": vulnerabilities,
        }

    def _security_scan_trivy(
        self,
        *,
        security_scan_config: GlobalSecurityScanConfig,
        repository: str,
        tag: str,
        log_image_ref: str,
        pull: bool,
    ) -> dict:
        # Pull image for scanning (if not already pulled) so that it can be scanned locally
        if pull:
            self._docker_client.pull(repository, tag)

        buildrunner_config = BuildRunnerConfig.get_instance()
        with tempfile.TemporaryDirectory(
            suffix="-trivy-run",
            dir=buildrunner_config.global_config.temp_dir,
        ) as local_run_dir:
            # Set constants for this run
            config_file_name = "config.yaml"
            results_file_name = "results.json"
            container_run_dir = "/trivy"
            # Dynamically use the configured cache directory if set, uses the trivy default otherwise
            container_cache_dir = security_scan_config.config.get(
                "cache-dir", "/root/.cache/trivy"
            )

            # Create local directories for volume mounting (if they don't exist)
            local_cache_dir = security_scan_config.cache_dir
            if not local_cache_dir:
                local_cache_dir = os.path.join(
                    buildrunner_config.global_config.temp_dir, "trivy-cache"
                )
            os.makedirs(local_cache_dir, exist_ok=True)

            # Create run config
            with open(
                os.path.join(local_run_dir, config_file_name), "w", encoding="utf8"
            ) as fobj:
                yaml.safe_dump(
                    {"cache-dir": container_cache_dir, **security_scan_config.config},
                    fobj,
                )

            image_scanner = None
            try:
                image_config = DockerRunner.ImageConfig(
                    f"{BuildRunnerConfig.get_instance().global_config.docker_registry}/"
                    f"aquasec/trivy:{security_scan_config.version}",
                    pull_image=False,
                )
                image_scanner = DockerRunner(
                    image_config,
                    log=self.step_runner.log,
                )
                image_scanner.start(
                    entrypoint="/bin/sh",
                    volumes={
                        local_run_dir: container_run_dir,
                        local_cache_dir: container_cache_dir,
                        # TODO Implement support for additional connection methods
                        "/var/run/docker.sock": "/var/run/docker.sock",
                    },
                )

                # Print out the trivy version
                image_scanner.run("trivy --version", console=self.step_runner.log)

                # Run trivy
                start_time = time.time()
                exit_code = image_scanner.run(
                    f"trivy --config {container_run_dir}/{config_file_name} image "
                    f"-f json -o {container_run_dir}/{results_file_name} {repository}:{tag}",
                    console=self.step_runner.log,
                )
                LOGGER.info(
                    f"Took {round(time.time() - start_time, 1)} second(s) to scan image"
                )
                if exit_code:
                    raise BuildRunnerProcessingError(
                        f"Could not scan {log_image_ref} with trivy, see errors above"
                    )

                # Load results file and parse the max score
                results_file = os.path.join(local_run_dir, results_file_name)
                if not os.path.exists(results_file):
                    raise BuildRunnerProcessingError(
                        f"Results file {results_file} from trivy for {log_image_ref} does not exist, "
                        "check for errors above"
                    )
                with open(results_file, "r", encoding="utf8") as fobj:
                    results = yaml.safe_load(fobj)
                if not results:
                    raise BuildRunnerProcessingError(
                        f"Could not read results file {results_file} from trivy for {log_image_ref}, "
                        "check for errors above"
                    )
                return self._security_scan_trivy_parse_results(
                    security_scan_config, results
                )
            finally:
                if image_scanner:
                    # make sure the current user/group ids of our
                    # process are set as the owner of the files
                    exit_code = image_scanner.run(
                        f"chown -R {int(os.getuid())}:{int(os.getgid())} {container_run_dir}",
                        log=self.step_runner.log,
                    )
                    if exit_code != 0:
                        LOGGER.error("Error running trivy--unable to change ownership")
                    image_scanner.cleanup()

    def run(self, context):  # pylint: disable=too-many-branches
        # Tag multi-platform images
        built_image = context.get("mp_built_image")
        if built_image:
            # These are used in the image artifacts below, and should match for all tagged images
            built_image_ids_str = ",".join([
                image.trunc_digest for image in built_image.built_images
            ])
            built_image_id_with_platforms = [
                f"{image.platform}:{image.trunc_digest}"
                for image in built_image.built_images
            ]

            for repo in self._repos:
                tagged_image = built_image.add_tagged_image(repo.repository, repo.tags)

                # Add tagged image refs to committed images for use in determining if pull should be true/false
                for image_ref in tagged_image.image_refs:
                    self.step_runner.build_runner.committed_images.add(image_ref)

                # Add tagged image as artifact if this is a push and not just a commit
                if not self._commit_only:
                    self.step_runner.build_runner.add_artifact(
                        repo.repository,
                        {
                            "type": "docker-image",
                            "docker:image": built_image_ids_str,
                            "docker:repository": repo.repository,
                            "docker:tags": repo.tags,
                            "docker:platforms": built_image_id_with_platforms,
                            **self._security_scan_mp(
                                built_image,
                                f"{repo.repository}:{repo.tags[0]}",
                                repo.security_scan,
                            ),
                        },
                    )

            # Tag all images locally for the native platform
            self.step_runner.multi_platform.tag_native_platform(built_image)

        # Tag single platform images
        else:
            # first see if a run task produced an image (via a post-build config)
            if "run-image" in context:
                image_to_use = context.get("run-image")
            # next see if there was a run task, committing the end state of the
            # container as the image to use
            elif "run_runner" in context:
                image_to_use = context["run_runner"].commit(self.step_runner.log)
            # finally see if we have an image from a build task
            else:
                image_to_use = context.get("image", None)

            # validate we have an image
            if not image_to_use:
                raise BuildRunnerProcessingError(
                    "Cannot find an image to tag/push from a previous task"
                )
            self.step_runner.log.write(f"Using image {image_to_use} for tagging\n")

            # add the image to the list of generated images for potential cleanup
            self.step_runner.build_runner.generated_images.append(image_to_use)

            for repo in self._repos:
                if self._commit_only:
                    self.step_runner.log.write(
                        f'Committing resulting image as "{repo.repository}" with tags {", ".join(repo.tags)}.\n'
                    )
                else:
                    self.step_runner.log.write(
                        f'Preparing resulting image for push to "{repo.repository}" with tags {", ".join(repo.tags)}.\n'
                    )

                # Tag the image
                for tag in repo.tags:
                    self._docker_client.tag(
                        image_to_use,
                        repo.repository,
                        tag=tag,
                        force=True,
                    )
                    self.step_runner.build_runner.committed_images.add(
                        f"{repo.repository}:{tag}"
                    )

                    if not self._commit_only:
                        self.step_runner.build_runner.repo_tags_to_push.append((
                            f"{repo.repository}:{tag}",
                            # Used to be insecure registry, but this is now deprecated/removed
                            False,
                        ))

                # add image as artifact
                if not self._commit_only:
                    self.step_runner.build_runner.add_artifact(
                        os.path.join(self.step_runner.name, image_to_use),
                        {
                            "type": "docker-image",
                            "docker:image": image_to_use,
                            "docker:repository": repo.repository,
                            "docker:tags": repo.tags,
                            **self._security_scan_single(
                                repo.repository, repo.tags[-1], repo.security_scan
                            ),
                        },
                    )


class CommitBuildStepRunnerTask(PushBuildStepRunnerTask):
    """
    Class used to commit the resulting image (either from the build task, or if
    there is a run task, the snapshot of the resulting run container) with a
    tag matching the given registry/repository.
    """

    def __init__(self, step_runner, commits: List[StepPushCommit]):
        # Subclasses the push task, just set commit only to true
        super().__init__(step_runner, commits, commit_only=True)



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import codecs
from typing import Optional

from ..models import GlobalConfig


def fetch_file(parsed_url, _: Optional[GlobalConfig]):
    """
    Pull files from the local file system.
    """
    with codecs.open(parsed_url.path, "r", encoding="utf-8") as _file:
        contents = "".join(_file.readlines())

    return contents



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from collections import OrderedDict
from datetime import datetime
import io
import logging
import os
import re
import sys
import uuid
import portalocker
import timeout_decorator
import yaml.resolver
import yaml.scanner
import glob
import hashlib
from typing import Optional, Tuple

from buildrunner import loggers
from buildrunner.errors import BuildRunnerConfigurationError


LOCK_TIMEOUT_SECONDS = 1800.0
LOGGER = logging.getLogger(__name__)


class FailureToAcquireLockException(Exception):
    """
    Raised when there is failure to acquire file lock
    """

    pass


class OrderedLoader(yaml.Loader):  # pylint: disable=too-many-ancestors
    """
    Custom loader class that preserves dictionary order.
    """

    pass


def construct_mapping(loader, node):
    """
    :param loader:
    :param node:
    """
    loader.flatten_mapping(node)
    return OrderedDict(loader.construct_pairs(node))


OrderedLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
)
# Tell YAML how to dump the OrderedDict
yaml.add_representer(
    OrderedDict,
    lambda dumper, data: dumper.represent_dict(iter(data.items())),
)
yaml.add_representer(
    OrderedDict,
    lambda dumper, data: dumper.represent_dict(iter(data.items())),
    Dumper=yaml.SafeDumper,
)


class IgnoreAliasesDumper(yaml.Dumper):  # pylint: disable=too-many-ancestors
    """
    Custom dumper class that removes aliases.
    """

    def ignore_aliases(self, data):
        return True


def load_config(stream, cfg_file):
    """
    Load yaml while preserving the order of attributes in maps/dictionaries and
    removing any aliases.
    """
    # run the data through pyyaml again to remove any aliases
    try:
        return yaml.load(
            yaml.dump(
                yaml.load(stream, OrderedLoader),
                default_flow_style=False,
                Dumper=IgnoreAliasesDumper,
            ),
            Loader=OrderedLoader,
        )
    except yaml.scanner.ScannerError as err:
        raise BuildRunnerConfigurationError(
            f"The {cfg_file} file contains malformed yaml, "
            f"please check the syntax and try again: {err}"
        ) from err


def sanitize_tag(tag):
    """
    Sanitize a tag to remove illegal characters.

    :param tag: The tag to sanitize.
    :return: The sanitized tag.
    """
    _tag = re.sub(r"[^-_\w.]+", "-", tag.lower())
    if _tag != tag:
        LOGGER.debug(
            f"Forcing tag to lowercase and removing illegal characters: {tag} => {_tag}"
        )
    return _tag


def is_dict(obj):
    """
    Determines whether an object acts like a dict.

    Args:
      obj (object): The object to test.

    Return:
      True if the obj acts like a dict, False otherwise.
    """
    return hasattr(obj, "keys") and hasattr(obj, "__getitem__")


def epoch_time():
    """Return the current epoch time in integer seconds."""
    return int((datetime.utcnow() - datetime.utcfromtimestamp(0)).total_seconds())


def tempfile(prefix=None, suffix=None, temp_dir="/tmp"):
    """
    Generate a temporary file path within the container.
    """
    name = str(uuid.uuid4())
    if suffix:
        name = name + suffix
    if prefix:
        name = prefix + name

    return os.path.join(temp_dir, name)


def checksum(*files: Tuple[str]) -> str:
    """
    Generate a single SHA1 checksum of the list of files passed in
    """
    if not isinstance(files, tuple):
        LOGGER.warning(
            f"TypeError - Input 'files' needs to be a tuple instead of {type(files)}"
        )
        sys.exit()

    blocksize = 2**16
    hasher = hashlib.sha1()

    for filename in sorted(files):
        if not os.path.isfile(filename):
            LOGGER.warning(f"{filename} does not exist, skipping file.")
            continue

        with open(filename, "rb") as open_file:
            buf = open_file.read(blocksize)
            while len(buf) > 0:
                hasher.update(buf)
                buf = open_file.read(blocksize)

    return hasher.hexdigest()


def hash_sha1(file_name_globs=None):
    """
    Return the sha1 hash from the content of multiple files, represented by a list of globs
    """
    if not file_name_globs:
        file_name_globs = []
    blocksize = 2**16  # 65,536
    hasher = hashlib.sha1()
    for file_name_glob in file_name_globs:
        for file_name in sorted(glob.glob(file_name_glob)):
            try:
                # Use blocksize to ensure python memory isn't too full
                with open(file_name, "rb") as open_file:
                    buf = open_file.read(blocksize)
                    while len(buf) > 0:
                        hasher.update(buf)
                        buf = open_file.read(blocksize)
            except Exception:  # pylint: disable=broad-except
                LOGGER.warning(f"Error reading file: {file_name}")
    return hasher.hexdigest()


def _acquire_flock_open(
    lock_file: str,
    logger: loggers.ContainerLogger,
    mode: str,
    timeout_seconds: float = LOCK_TIMEOUT_SECONDS,
    exclusive: bool = True,
) -> io.IOBase:
    """
    Acquire file lock and open file with configurable timeout

    :param lock_file: path and file name of file open and lock
    :param logger: logger to log messages
    :param mode: mode used by open()
    :param timeout_seconds: number of seconds for timeout
    :param exclusive: config exclusive lock (True) or shared lock (False), defaults to True
    :return: opened file object if successful else None
    """

    @timeout_decorator.timeout(
        seconds=timeout_seconds, timeout_exception=FailureToAcquireLockException
    )
    def get_lock(file_obj: io.IOBase):
        flags = (
            portalocker.LockFlags.EXCLUSIVE
            if exclusive
            else portalocker.LockFlags.SHARED
        )
        portalocker.lock(
            file_obj,
            flags,
        )
        return file_obj

    # pylint: disable=unspecified-encoding,consider-using-with
    file_obj = open(lock_file, mode)
    lock_file_obj = None
    pid = os.getpid()

    try:
        lock_file_obj = get_lock(file_obj)
    except FailureToAcquireLockException:
        file_obj.close()
        raise FailureToAcquireLockException(
            f"PID:{pid} failed to acquire file lock for {lock_file} after timeout of {timeout_seconds} seconds"
        )

    logger.info(f"PID:{pid} file opened and lock acquired for {lock_file}")

    return lock_file_obj


def acquire_flock_open_read_binary(
    lock_file: str,
    logger: loggers.ContainerLogger,
    timeout_seconds: float = LOCK_TIMEOUT_SECONDS,
) -> io.BufferedReader:
    """
    Acquire file lock and open binary file in read mode with configurable timeout

    :param lock_file: path and file name of file open and lock
    :param logger: logger to log messages
    :param timeout_seconds: number of seconds for timeout
    :return: opened file object
    """
    return _acquire_flock_open(
        lock_file=lock_file,
        logger=logger,
        mode="rb",
        timeout_seconds=timeout_seconds,
        exclusive=False,
    )


def acquire_flock_open_write_binary(
    lock_file: str,
    logger: loggers.ContainerLogger,
    timeout_seconds: float = LOCK_TIMEOUT_SECONDS,
    mode: str = "wb",
) -> io.BufferedWriter:
    """
    Acquire file lock and open binary file in write mode with configurable timeout

    :param lock_file: path and file name of file open and lock
    :param logger: logger to log messages
    :param timeout_seconds: number of seconds for timeout
    :param exclusive: config exclusive lock (True) or shared lock (False), defaults to True
    :return: opened file object
    """
    return _acquire_flock_open(
        lock_file=lock_file,
        logger=logger,
        mode=mode,
        timeout_seconds=timeout_seconds,
        exclusive=True,
    )


def release_flock(
    lock_file_obj: Optional[io.BufferedReader], logger: loggers.ContainerLogger
):
    """
    Release the file lock and close file descriptor

    :param lock_file_obj: opened lock file object
    :param logger: logger to log messages
    """
    if lock_file_obj is None:
        return
    portalocker.unlock(lock_file_obj)
    lock_file_obj.close()
    logger.info(f"PID:{os.getpid()} released and closed file {lock_file_obj.name}")



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import yaml

import docker
import docker.errors

from buildrunner.docker import new_client
from buildrunner.errors import BuildRunnerProcessingError
from buildrunner.utils import is_dict


class DockerImporter:
    """
    An object that orchestrates importing a Docker image from
    a tar file.
    """

    def __init__(
        self,
        src,
        dockerd_url=None,
        timeout=None,
    ):
        self.src = src
        self.docker_client = new_client(
            dockerd_url=dockerd_url,
            timeout=timeout,
        )
        self.image = None

    def import_image(self):
        """
        Run a docker import using the configured src archive.
        """

        try:
            import_return = self.docker_client.import_image(self.src)
        except docker.errors.APIError as apie:
            raise BuildRunnerProcessingError(
                f"Error importing image from archive file {self.src}: {apie}"
            ) from apie
        if not is_dict(import_return):
            import_return = yaml.safe_load(import_return)
        if "status" not in import_return:
            raise BuildRunnerProcessingError(
                f"Error importing image from archive file {self.src}"
            )
        return import_return["status"]



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from buildrunner.errors import BuildRunnerProvisionerError

from buildrunner.provisioners.salt import SaltProvisioner
from buildrunner.provisioners.shell import ShellProvisioner

PROVISIONERS = {
    "shell": ShellProvisioner,
    "salt": SaltProvisioner,
}


def create_provisioners(provisioners_config: dict, logger):
    """
    Given provisioners config return a list of provisioner objects.
    """
    _provisioners = []

    if provisioners_config:
        for _type, _value in provisioners_config.items():
            if _type in PROVISIONERS:
                _provisioners.append(PROVISIONERS[_type](_value, console=logger))
            else:
                raise BuildRunnerProvisionerError(f'Unknown provisioner type "{_type}"')

    return _provisioners



"""
Copyright 2023 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from buildrunner.config.models_step import StepTask


class BuildStepRunnerTask:
    """
    Base task class.
    """

    def __init__(self, step_runner, step: StepTask):
        """
        Subclasses should extend this method to validate the task configuration
        and set any attributes.
        """
        self.step_runner = step_runner
        self.step = step

    def run(self, context):
        """
        Subclasses override this method to perform their task.
        """
        pass

    def cleanup(self, context):
        """
        Subclasses override this method to perform any cleanup tasks.
        """
        pass



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
import io
import struct
import threading
import time
import urllib.parse

from paramiko import (
    DSSKey,
    MissingHostKeyPolicy,
    PasswordRequiredException,
    RSAKey,
    ECDSAKey,
    Ed25519Key,
    SSHClient,
    SSHException,
)
from paramiko.agent import AgentSSH
from paramiko.common import io_sleep
from paramiko.util import asbytes
from paramiko.message import Message

import buildrunner.docker.builder
from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProcessingError,
)

SSH_AGENT_PROXY_BUILD_CONTEXT = os.path.join(
    os.path.dirname(__file__), "SSHAgentProxyImage"
)
SSH_KEY_TYPES = [RSAKey, ECDSAKey, Ed25519Key, DSSKey]


def load_ssh_key_from_file(key_file, passwd):
    """
    Load the given keys into paramiko PKey objects.
    """
    for key_type in SSH_KEY_TYPES:
        try:
            return key_type.from_private_key_file(key_file, passwd)
        except PasswordRequiredException as pwdreqe:
            raise BuildRunnerConfigurationError(
                f"Key at {key_file} requires a password"
            ) from pwdreqe
        except SSHException:
            continue
    raise BuildRunnerConfigurationError(f"Unable to load key at {key_file}")


def load_ssh_key_from_str(key_str, passwd):
    """
    Load the given keys into paramiko PKey objects.
    """
    try:
        return RSAKey.from_private_key(io.StringIO(key_str), passwd)
    except PasswordRequiredException as pwdreqe:
        raise BuildRunnerConfigurationError(
            "Provided key requires a password"
        ) from pwdreqe
    except SSHException:
        try:
            return DSSKey.from_private_key(
                io.StringIO(key_str),
                passwd,
            )
        except PasswordRequiredException as pwdreqe:
            raise BuildRunnerConfigurationError(
                "Provided key requires a password"
            ) from pwdreqe
        except SSHException as sshe:
            raise BuildRunnerConfigurationError("Unable to load provided key") from sshe


class DockerSSHAgentProxy:
    """
    Class used to manage a Docker container that exposes a ssh-agent socket
    through a volume that can be mounted within other containers running within
    the same Docker host. Keys are loaded and shared through a custom ssh-agent
    implementation that is managed by this class.
    """

    def __init__(
        self,
        docker_client,
        log,
        docker_registry,
        multiplatform_image_builder,
        container_labels,
        network,
    ):
        """ """
        self.docker_client = docker_client
        self.log = log
        self.docker_registry = docker_registry
        self._ssh_agent_image = None
        self._ssh_agent_container = None
        self._ssh_client = None
        self._ssh_channel = None
        self._multiplatform_image_builder = multiplatform_image_builder
        self._container_labels = container_labels
        self._network = network

    def get_info(self):
        """
        Return a tuple where the first item is the ssh-agent container id and
        the second is a dict of environment variables to be injected into other
        containers providing settings for ssh clients to connect to the shared
        agent.
        """
        return (self._ssh_agent_container, {"SSH_AUTH_SOCK": "/ssh-agent/agent"})

    def start(self, keys):
        """
        Loads the given keys, starts a Docker container, creates a persistant
        ssh connection to the container, and starts the custom ssh agent
        thread.

        Args:
          - keys a dict with the key being the file path and the value being a
            password (or null if not required)
        """

        # load the keys
        if not keys:
            raise BuildRunnerConfigurationError("Invalid private keys")

        # create and start the Docker container
        self._ssh_agent_container = self.docker_client.create_container(
            self.get_ssh_agent_image(),
            command=[
                f"{keys[0].get_name()} {keys[0].get_base64()}",
            ],
            labels=self._container_labels,
            networking_config=self.docker_client.create_networking_config({
                self._network: self.docker_client.create_endpoint_config()
            })
            if self._network
            else None,
            host_config=self.docker_client.create_host_config(
                publish_all_ports=True,
            ),
        )["Id"]
        self.docker_client.start(
            self._ssh_agent_container,
        )
        self.log.write(f"Created ssh-agent container {self._ssh_agent_container:.10}\n")

        _ssh_host = "localhost"
        # See if buildrunner is executing from a container.  If so, hit the
        # newly created container directly on port 22
        if os.environ.get("BUILDRUNNER_CONTAINER"):
            _ssh_container = self.docker_client.inspect_container(
                self._ssh_agent_container
            )
            network_settings = _ssh_container.get("NetworkSettings", {})
            if self._network:
                _ssh_host = (
                    network_settings.get("Networks", {})
                    .get(self._network, {})
                    .get("IPAddress", _ssh_host)
                )
            else:
                _ssh_host = network_settings.get("IPAddress", _ssh_host)
            _ssh_port = 22
        else:
            # get the Docker server ip address and ssh port exposed by this
            # container
            p_data = urllib.parse.urlparse(self.docker_client.base_url)
            if p_data and "unix" not in p_data.scheme and p_data.hostname:
                if p_data.hostname != "localunixsocket":
                    _ssh_host = p_data.hostname
            _ssh_port_info = self.docker_client.port(
                self._ssh_agent_container,
                22,
            )
            if not _ssh_port_info or "HostPort" not in _ssh_port_info[0]:
                raise BuildRunnerProcessingError(
                    "Unable to find port for ssh-agent container"
                )
            _ssh_port = _ssh_port_info[0]["HostPort"]
            _ssh_port = int(_ssh_port)
        time.sleep(3)

        # setup ssh connection with fake agent in own thread
        self._ssh_client = SSHClient()
        self._ssh_client.set_missing_host_key_policy(MissingHostKeyPolicy())
        # pylint: disable=W0212
        self._ssh_client._agent = CustomSSHAgent(keys)

        self._try_connect(_ssh_host, _ssh_port)

        self._ssh_channel = self._ssh_client.get_transport().open_session()
        # AgentRequestHandler(channel)
        self._ssh_channel.request_forward_agent(
            self._ssh_client._agent.forward_agent_handler
        )
        self._ssh_channel.get_pty()
        self._ssh_channel.exec_command("/login.sh")
        self.log.write("Established ssh-agent container connection\n")

    def _try_connect(self, ssh_host, ssh_port):
        """
        Try to connect ssh_client with a retry/backoff. The retry is to help deal with situations where the sshd process
        may not yet be ready. Retrying with a backoff gives it a chance before we give up for good
        """
        previous_backoff = 0
        backoff = 1
        while backoff <= 8:
            try:
                self._ssh_client.connect(
                    ssh_host,
                    port=ssh_port,
                    username="root",
                    allow_agent=True,
                    look_for_keys=False,
                )
                break
            except Exception as exc:  # pylint: disable=broad-except
                self.log.write(
                    f"There was an issue trying to connect to container: {exc}"
                )
            next_backoff = backoff + previous_backoff
            previous_backoff = backoff
            backoff = next_backoff
            time.sleep(backoff)
        else:
            self.log.write(f"Unable to successfully connect to {ssh_host}\n")
            # pylint: disable=broad-exception-raised
            raise Exception(f"Unable to successfully connect to {ssh_host}")

    def stop(self):
        """
        Stops the custom agent thread, kills the persistent ssh connection to
        the remote container, and kills the container.
        """
        # kill ssh connection thread
        self.log.write("Closing ssh-agent container connection\n")
        if self._ssh_client:
            if self._ssh_client._agent:
                try:
                    self._ssh_client._agent.close()
                except Exception as _ex:
                    self.log.write(f"Error stopping ssh-agent: {_ex}\n")
            try:
                self._ssh_client.close()
            except Exception as _ex:
                self.log.write(f"Error stopping ssh-agent connection: {_ex}\n")

        # kill container
        if self._ssh_agent_container:
            self.log.write(
                f"Destroying ssh-agent container {self._ssh_agent_container:.10}\n"
            )
            try:
                self.docker_client.remove_container(
                    self._ssh_agent_container,
                    force=True,
                    v=True,
                )
            except Exception as _ex:
                self.log.write(f"Error destroying ssh-agent container: {_ex}\n")

    def get_ssh_agent_image(self):
        """
        Get and/or create the image used to proxy the ssh agent to a container.
        """
        if not self._ssh_agent_image:
            self.log.write("Creating ssh-agent image\n")
            image = buildrunner.docker.builder.build_image(
                path=SSH_AGENT_PROXY_BUILD_CONTEXT,
                docker_registry=self.docker_registry,
                nocache=False,
                pull=False,
            )
            self._ssh_agent_image = image
        return self._ssh_agent_image


class CustomSSHAgent(AgentSSH):
    """
    Custom class implementing the paramiko ssh agent apis.
    """

    def __init__(self, keys):
        AgentSSH.__init__(self)
        self._conn = None
        self._keys = keys
        self._connection_threads = []

    def _connect(self, conn):
        """
        Override parent.
        """
        pass

    def __del__(self):
        self.close()

    def close(self):
        """
        Override parent.
        """
        if self._connection_threads:
            for _ct in self._connection_threads:
                _ct.stop()
        self._keys = []

    def get_keys(self):
        """
        Return the keys.
        """
        return tuple(self._keys)

    def forward_agent_handler(self, remote_channel):
        """
        Handler function for setting up the thread that handles remote ssh
        agent requests from the server we connect to.
        """
        _ct = CustomAgentConnectionThread(
            self,
            remote_channel,
        )
        self._connection_threads.append(_ct)
        _ct.start()


SSH2_AGENT_FAILURE = bytes([30])
SSH2_AGENTC_REQUEST_IDENTITIES = 11
SSH2_AGENT_IDENTITIES_ANSWER = bytes([12])
SSH2_AGENTC_SIGN_REQUEST = 13
SSH2_AGENT_SIGN_RESPONSE = bytes([14])


class CustomAgentConnectionThread(threading.Thread):
    """
    Class that manages a remote (upstream server) connection to a
    CustomSSHAgent.
    """

    def __init__(self, agent, remote_channel):
        threading.Thread.__init__(self, target=self.run)
        self._agent = agent
        self._remote_channel = remote_channel
        self._exit = False

    def run(self):
        """
        Main server routine.
        """
        try:
            while not self._exit:
                try:
                    if self._remote_channel.eof_received:
                        self.stop()
                        continue

                    if self._remote_channel.recv_ready():
                        r_type, request = self._receive_request()

                        if r_type == SSH2_AGENTC_REQUEST_IDENTITIES:
                            self._agent_identities_answer()
                        elif r_type == SSH2_AGENTC_SIGN_REQUEST:
                            self._agent_sign_response(request)
                        else:
                            # return FAILURE message for everything else
                            self._send_reply(SSH2_AGENT_FAILURE)
                except SSHException:
                    raise
                except Exception:  # pylint: disable=broad-except
                    pass
                time.sleep(io_sleep)
        # pylint: disable=W0703
        except Exception:
            self.stop()

    def _agent_identities_answer(self):
        """
        Return the keys in the custom agent.
        """
        msg = Message()
        msg.add_byte(SSH2_AGENT_IDENTITIES_ANSWER)
        _keys = self._agent.get_keys()
        msg.add_int(len(_keys))
        for _key in _keys:
            # add each key
            msg.add_string(_key.asbytes())
            msg.add_string("")
        self._send_reply(msg)

    def _agent_sign_response(self, request):
        """
        Lookup the key in the custom agent and use it to sign the passed data.
        """
        key_blob = request.get_string()
        data = request.get_string()

        signed = None
        for _key in self._agent.get_keys():
            if key_blob == _key.asbytes():
                _signed = _key.sign_ssh_data(data)
                signed = _signed.asbytes()
        msg = Message()
        if signed:
            msg.add_byte(SSH2_AGENT_SIGN_RESPONSE)
            msg.add_string(signed)
        else:
            msg.add_byte(SSH2_AGENT_FAILURE)
        self._send_reply(msg)

    def _send_reply(self, msg):
        """
        Send a reply back to the upstream agent.
        """
        raw_msg = asbytes(msg)
        length = struct.pack(">I", len(raw_msg))
        self._remote_channel.send(length + raw_msg)

    def _receive_request(self):
        """
        Receive the request, storing the bytes in a Message object for easy
        retrieval of message parts.
        """
        message_length = self._read_all(4)
        msg = Message(self._read_all(struct.unpack(">I", message_length)[0]))
        return ord(msg.get_byte()), msg

    def _read_all(self, wanted):
        """
        Read the specified bytes from the remote channel.
        """
        result = self._remote_channel.recv(wanted)
        while len(result) < wanted:
            if len(result) == 0:
                raise SSHException("lost upstream ssh-agent connection")
            extra = self._remote_channel.recv(wanted - len(result))
            if len(extra) == 0:
                raise SSHException("lost upstream ssh-agent connection")
            result += extra
        return result

    def stop(self):
        """
        Close the remote connection and stop the thread.
        """
        self._exit = True
        if self._remote_channel:
            self._remote_channel.close()
            self._remote_channel = None



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import base64
from collections import OrderedDict
import copy
import functools
import getpass
from graphlib import TopologicalSorter
from io import StringIO
import logging
import os
from typing import List, Optional, Union

import jinja2

from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerVersionError,
    ConfigVersionFormatError,
    ConfigVersionTypeError,
)
from buildrunner.utils import (
    checksum,
    hash_sha1,
    load_config,
)

from .models import GlobalConfig

from . import fetch, jinja_context


MASTER_GLOBAL_CONFIG_FILE = "/etc/buildrunner/buildrunner.yaml"
RESULTS_DIR = "buildrunner.results"
LOGGER = logging.getLogger(__name__)


def _log_generated_file(
    log_generated_files: bool, file_name: str, file_contents: str
) -> None:
    """
    Conditionally log the contents of a generated file.
    """
    if log_generated_files:
        LOGGER.info(f"Generated contents of {file_name}:")
        for line in file_contents.splitlines():
            LOGGER.info(line)


def _add_default_tag_to_tags(config: Union[str, dict], default_tag: str) -> dict:
    # Add default tag to tags list if not in the list already
    if isinstance(config, dict):
        if not config.get("tags"):
            config["tags"] = []
        assert isinstance(config.get("tags"), list)
        # Only add the tag if not disabled in the config
        if config.get("add_build_tag", True) and default_tag not in config.get("tags"):
            config.get("tags").append(default_tag)

    # Convert to dictionary and add default tag if not listed already
    elif isinstance(config, str):
        image_name = config.split(":")
        step_config = dict()
        step_config["repository"] = image_name[0]
        step_config["tags"] = []

        # Check if image name has defined a tag, if so add it to tags
        if len(image_name) > 1:
            tag = image_name[1]
            assert tag
            step_config["tags"].append(tag)

        # Add default tag if not in the list already
        if default_tag not in step_config["tags"]:
            step_config["tags"].append(default_tag)

        config = step_config
    return config


def _set_default_tag(config: dict, default_tag: str) -> dict:
    """
    Set default tag if not set for each image

    Args:
        config (dict):  configuration
        default_tag (str): default tag

        Returns:
            dict: configuration
    """
    steps = config.get("steps")
    if not isinstance(steps, dict):
        return config
    for step_name, step in steps.items():
        for substep_name, substep in step.items():
            if substep_name in ["push", "commit"]:
                # Add default tag to tags list if not in the list
                if isinstance(substep, list):
                    curr_image_infos = []
                    for push_config in substep:
                        curr_image_infos.append(
                            _add_default_tag_to_tags(push_config, default_tag)
                        )
                    config["steps"][step_name][substep_name] = curr_image_infos
                else:
                    curr_image_info = _add_default_tag_to_tags(substep, default_tag)
                    config["steps"][step_name][substep_name] = curr_image_info
    return config


def _validate_version(config: dict) -> None:
    """
    Compares that the version in the config is less than or equal to the current version of
    buildrunner. If the config version is greater than the buildrunner version or any parsing error occurs
    it will raise a buildrunner exception.
    """
    buildrunner_version = None

    from buildrunner import __version__

    parts = __version__.strip().split(".")
    buildrunner_version = f"{parts[0]}.{parts[1]}" if len(parts) >= 2 else None

    if not buildrunner_version:
        if __version__ != "DEVELOPMENT":
            raise BuildRunnerVersionError("unable to determine buildrunner version")
        LOGGER.warning(
            "Unable to determine buildrunner version for validation. Skipping config version check."
        )
        return

    # version is optional and is valid to not have it in the config
    if "version" not in config:
        return

    config_version = config["version"]

    try:
        if float(config_version) > float(buildrunner_version):
            raise ConfigVersionFormatError(
                f"configuration version {config_version} is higher than "
                f"buildrunner version {buildrunner_version}"
            )
    except ValueError as exception:
        raise ConfigVersionTypeError(
            f'unable to convert config version "{config_version}" '
            f'or buildrunner version "{buildrunner_version}" '
            f"to a float"
        ) from exception


def _reorder_dependency_steps(config: dict) -> dict:
    """
    Reorders the steps based on the dependencies that are outlined in the config
    """
    # Defines configuration keywords, should add to a config validation class
    keyword_version = "version"
    keyword_steps = "steps"
    keyword_depends = "depends"
    supported_version = 2.0

    if (
        keyword_version not in config.keys()
        or config[keyword_version] < supported_version
    ):
        return config

    ordered_steps = OrderedDict()
    if keyword_steps in config.keys():
        topo_sorter = TopologicalSorter()
        for name, instructions in config[keyword_steps].items():
            if keyword_depends in instructions.keys():
                for depend in instructions[keyword_depends]:
                    topo_sorter.add(name, depend)
            else:
                topo_sorter.add(name)
        for step in topo_sorter.static_order():
            if step not in config[keyword_steps]:
                raise KeyError(
                    f"Step '{step}' is not defined and is listed as a step dependency in "
                    f"the configuration. "
                    f"Please correct the typo or define step '{step}' in the configuration."
                )

            if keyword_depends in config[keyword_steps][step].keys():
                del config[keyword_steps][step][keyword_depends]

            ordered_steps[step] = config[keyword_steps][step]

        config[keyword_steps] = ordered_steps

    return config


def _fetch_template(
    *,
    env: dict,
    build_time: int,
    cfg_file: str,
    global_config: Optional[GlobalConfig] = None,
    ctx: Optional[dict] = None,
    log_file: bool = True,
) -> dict:
    """
    Load a config file templating it with Jinja and parsing the YAML.

    Returns:
        Dictionary parsed from the contents of the final loaded file
    """

    fetch_file = cfg_file
    visited = set()

    while True:
        visited.add(fetch_file)
        contents = fetch.fetch_file(fetch_file, global_config)
        jenv = jinja2.Environment(
            loader=jinja2.FileSystemLoader("."), extensions=["jinja2.ext.do"]
        )
        jenv.filters["hash_sha1"] = hash_sha1
        jenv.filters["base64encode"] = base64.encode
        jenv.filters["base64decode"] = base64.decode
        jenv.filters["re_sub"] = jinja_context.re_sub_filter
        jenv.filters["re_split"] = jinja_context.re_split_filter

        jenv.globals.update(checksum=checksum)
        jtemplate = jenv.from_string(contents)

        config_context = copy.deepcopy(env)
        config_context.update({
            "CONFIG_FILE": cfg_file,
            "CONFIG_DIR": os.path.dirname(cfg_file),
            "read_yaml_file": functools.partial(
                jinja_context.read_yaml_file, env, _log_generated_file, log_file
            ),
            "raise": jinja_context.raise_exception_jinja,
            "strftime": functools.partial(jinja_context.strftime, build_time),
            "env": os.environ,
            # This is stored after the initial env is set
            "DOCKER_REGISTRY": global_config.docker_registry if global_config else None,
        })

        if ctx:
            config_context.update(ctx)

        config_contents = jtemplate.render(config_context)
        _log_generated_file(log_file, cfg_file, config_contents)
        config = load_config(StringIO(config_contents), cfg_file)

        if not config:
            break

        redirect = config.get("redirect")
        if redirect is None:
            break

        fetch_file = redirect
        if fetch_file in visited:
            raise BuildRunnerConfigurationError(
                f"Redirect loop visiting previously visited file: {fetch_file}"
            )

    return config or {}


def load_run_file(
    *,
    global_config: GlobalConfig,
    env: dict,
    build_time: int,
    run_config_file: str,
    log_file: bool,
    default_tag: Optional[str] = None,
) -> dict:
    config_data = _fetch_template(
        global_config=global_config,
        env=env,
        build_time=build_time,
        cfg_file=run_config_file,
        log_file=log_file,
    )
    # Validate the version
    _validate_version(config_data)
    # Reorder steps for dependencies
    config_data = _reorder_dependency_steps(config_data)
    # Always add default tag if not set
    config_data = _set_default_tag(config_data, default_tag)

    return config_data


def _deep_merge_dicts(a_dict: dict, b_dict: dict, path=None) -> dict:
    if path is None:
        path = []
    for key in b_dict:
        if key in a_dict:
            if isinstance(a_dict[key], dict) and isinstance(b_dict[key], dict):
                _deep_merge_dicts(a_dict[key], b_dict[key], path + [str(key)])
            elif a_dict[key] != b_dict[key]:
                a_dict[key] = b_dict[key]
        else:
            a_dict[key] = b_dict[key]
    return a_dict


def load_global_config_files(
    *,
    build_time: int,
    global_config_files: List[str],
    global_config_overrides: dict,
) -> dict:
    """
    Load global config files templating them with Jinja and parsing the YAML.

    Returns:
      A dictionary of configuration
    """
    username = getpass.getuser()
    homedir = os.path.expanduser("~")

    context = {}
    for cfg in global_config_files:
        cfg_path = os.path.realpath(os.path.expanduser(cfg))
        if os.path.exists(cfg_path):
            current_context = _fetch_template(
                env={},
                build_time=build_time,
                cfg_file=cfg_path,
                ctx=context,
                log_file=False,
            )
            if current_context is None:
                # Empty config file
                continue

            # Only allow MASTER_GLOBAL_CONFIG_FILE to specify arbitrary local-files for mounting
            # - all other local-files get scrubbed for specific requirements and non-matches
            # are dropped.
            if cfg_path != MASTER_GLOBAL_CONFIG_FILE:
                scrubbed_local_files = {}
                for fname, fpath in list(
                    current_context.get("local-files", {}).items()
                ):
                    if not isinstance(fpath, str):
                        LOGGER.info(f'Bad "local-files" entry in {cfg_path!r}:')
                        LOGGER.info(f"    {fname!r}: {fpath!r}")
                        continue
                    resolved_path = os.path.realpath(os.path.expanduser(fpath))
                    # pylint: disable=too-many-boolean-expressions
                    if (
                        username == "root"
                        or resolved_path == homedir
                        or resolved_path.startswith(homedir + os.path.sep)
                        or os.stat(resolved_path).st_uid == os.getuid()
                        or (
                            not os.path.isdir(resolved_path)
                            and os.access(resolved_path, os.R_OK | os.W_OK)
                        )
                    ):
                        scrubbed_local_files[fname] = resolved_path
                    else:
                        LOGGER.info(f'Bad "local-files" entry in {cfg_path!r}:')
                        LOGGER.info(
                            f"    User {username!r} is not allowed to mount {resolved_path!r}."
                        )
                        LOGGER.info(
                            f"    You may need an entry in {MASTER_GLOBAL_CONFIG_FILE!r}."
                        )
                current_context["local-files"] = scrubbed_local_files

            _deep_merge_dicts(context, current_context)

    return _deep_merge_dicts(context, global_config_overrides)



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
import traceback
import uuid

from buildrunner.docker.multiplatform_image_builder import MultiplatformImageBuilder
from buildrunner.config.models_step import Step, StepBuild, StepRun
from buildrunner.errors import BuildRunnerError
from buildrunner.steprunner.tasks.build import BuildBuildStepRunnerTask
from buildrunner.steprunner.tasks.push import (
    CommitBuildStepRunnerTask,
    PushBuildStepRunnerTask,
)
from buildrunner.steprunner.tasks.pypipush import PypiPushBuildStepRunnerTask
from buildrunner.steprunner.tasks.remote import RemoteBuildStepRunnerTask
from buildrunner.steprunner.tasks.run import RunBuildStepRunnerTask

TASK_MAPPINGS = {
    "remote": RemoteBuildStepRunnerTask,
    "build": BuildBuildStepRunnerTask,
    "run": RunBuildStepRunnerTask,
    "push": PushBuildStepRunnerTask,
    "commit": CommitBuildStepRunnerTask,
    "pypi-push": PypiPushBuildStepRunnerTask,
}


class BuildStepRunner:  # pylint: disable=too-many-instance-attributes
    """
    Class used to manage running a build step.
    """

    class ImageConfig:
        """
        An object that captures image-specific configuration
        """

        def __init__(self, local_images=False, platform=None):
            self.local_images = local_images
            self.platform = platform

    def __init__(
        self,  # pylint: disable=too-many-arguments
        build_runner,
        step_name,
        step: Step,
        image_config,
        multi_platform: MultiplatformImageBuilder,
        container_labels,
    ):
        """
        Constructor.
        """
        local_images = image_config.local_images
        platform = image_config.platform
        self.name = step_name
        self.step = step
        self.local_images = local_images
        self.platform = platform

        self.build_runner = build_runner
        self.src_dir = self.build_runner.build_dir
        self.results_dir = os.path.join(
            self.build_runner.build_results_dir,
            self.name,
        )
        if not os.path.exists(self.results_dir):
            os.mkdir(self.results_dir)

        self.log = self.build_runner.log

        # generate a unique step id
        self.id = str(uuid.uuid4())  # pylint: disable=invalid-name
        self.multi_platform = multi_platform
        self.container_labels = container_labels
        # network name is used to identify the network that the build step is running in
        self.network_name = f"{self.id}-network"

    def run(self):
        """
        Run the build step.
        """
        # create the step results dir
        self.log.write(f'\nRunning step "{self.name}"\n')
        self.log.write(f"{'_' * 40}\n")

        _tasks = []
        _context = {}
        try:
            for task_name, task_class in TASK_MAPPINGS.items():
                task_data = getattr(self.step, task_name.replace("-", "_"))
                if not task_data:
                    continue
                self.log.write(f"==> Running step: {self.name}:{task_name}\n")
                if isinstance(task_data, (StepBuild, StepRun)):
                    if self.local_images:
                        task_data.pull = False
                    if self.platform:
                        task_data.platform = self.platform
                _task = task_class(self, task_data)
                _tasks.append(_task)
                try:
                    _task.run(_context)
                except BuildRunnerError as err:
                    if not isinstance(task_data, StepRun) or not task_data.xfail:
                        raise

                    self.log.write(
                        f'Step "{self.name}" failed with exception: {err}\n    '
                        f"Ignoring due to XFAIL\n"
                    )
                except Exception as err:  # pylint: disable=broad-except
                    self.log.write(f'Step "{self.name}" failed with exception: {err}\n')
                    raise err
        finally:
            for _task in _tasks:
                try:
                    _task.cleanup(_context)
                except Exception:  # pylint: disable=broad-except
                    self.log.write("\nError cleaning up task:\n")
                    traceback.print_exc(file=self.log)



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import logging
import os
import sys
import queue
from typing import Optional, Union

import colorlog
from rich import progress


CONSOLE_LOGGER_NAME = "buildrunner"
ENCODING = sys.stdout.encoding if sys.stdout.encoding else "utf8"


class CustomColoredFormatter(colorlog.ColoredFormatter):
    def __init__(self, fmt: str, no_color: bool, color: str = "white"):
        super().__init__(
            fmt, no_color=no_color, log_colors=self._get_log_level_colors(color)
        )
        self.fmt = fmt
        self.color = color

    def clone(
        self, color: Optional[str] = None, no_color: Optional[bool] = None
    ) -> "CustomColoredFormatter":
        if color is None:
            color = self.color
        if no_color is None:
            no_color = self.no_color
        return CustomColoredFormatter(self.fmt, no_color, color)

    @staticmethod
    def _get_log_level_colors(color: str) -> dict:
        return {
            "DEBUG": color,
            "INFO": color,
            "WARNING": color,
            "ERROR": color,
            "CRITICAL": color,
        }


def get_build_log_file_path(build_results_dir: str) -> str:
    return os.path.join(build_results_dir, "build.log")


def _get_logger_format(no_log_color: bool, disable_timestamps: bool):
    timestamp = "" if disable_timestamps else "%(asctime)s "
    return CustomColoredFormatter(
        f"%(log_color)s{timestamp}%(levelname)-8s %(message)s",
        no_log_color,
    )


def initialize_root_logger(
    debug: bool, no_log_color: bool, disable_timestamps: bool, build_results_dir: str
) -> None:
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG if debug else logging.INFO)

    console_formatter = _get_logger_format(no_log_color, disable_timestamps)
    # The file formatter should always use no color and timestamps should be enabled
    file_formatter = _get_logger_format(True, False)

    console_handler = logging.StreamHandler(sys.stdout)
    # Console logger should use colored output when specified by config
    console_handler.setFormatter(console_formatter)
    file_handler = logging.FileHandler(
        get_build_log_file_path(build_results_dir), "w", encoding="utf8"
    )
    # The build log should never use colored output
    file_handler.setFormatter(file_formatter)
    logger.handlers.clear()
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)


class ConsoleLogger:
    """
    Class inherited from logger that provides backwards support for the "write" method.
    This should be removed at some point and reimplemented with actual logging calls.
    """

    def __init__(self, name: str):
        self.logger = logging.getLogger(name)

    @staticmethod
    def clean_output(output: Union[bytes, str]) -> str:
        if not isinstance(output, str):
            output = str(output, encoding=ENCODING, errors="replace")
        if output and output[-1] == "\n":
            output = output[:-1]
        return output

    def write(self, output: Union[bytes, str]):
        """
        Write the given text to stdout and streams decorating output to stdout
        with color.
        """
        for line in self.clean_output(output).split("\n"):
            self.logger.info(line)

    # Delegates all methods to the logger if they don't exist (allowing the logger methods to be used directly)
    def __getattr__(self, item):
        return getattr(self.logger, item)


class ColorQueue(queue.SimpleQueue):
    def __init__(self, *colors):
        super().__init__()
        for color in colors:
            self.put(color)

    def next(self) -> str:
        current = self.get()
        self.put(current)
        return current


class ContainerLogger(ConsoleLogger):
    """
    Class used to write container specific output to logging.

    This class is not thread safe, but since each container gets its own that
    is ok.
    """

    BUILD_LOG_COLORS = ColorQueue("yellow", "blue")
    SERVICE_LOG_COLORS = ColorQueue("purple", "cyan", "red", "green")
    LOGGERS = {}

    def __init__(self, name: str, color: str, prefix: Optional[str] = None):
        super().__init__(name)
        self._set_logger_handlers(color)

        self._line_prefix = f"[{prefix if prefix else name}] "
        self._buffer = []

    def _set_logger_handlers(self, color: str) -> None:
        """
        Copy and modify the handlers from the root logger in order to change the color for the console logger
        as well as push the logs to the results build log file
        :param color: the color to set on the console handler
        """
        # Only override this if not already set
        if self.logger.handlers:
            return

        self.logger.propagate = False
        for handler in logging.getLogger().handlers:
            if (
                isinstance(handler, logging.StreamHandler)
                and not isinstance(handler, logging.FileHandler)
                and isinstance(handler.formatter, CustomColoredFormatter)
            ):
                new_handler = logging.StreamHandler(handler.stream)
                new_handler.setFormatter(handler.formatter.clone(color))
                self.logger.addHandler(new_handler)
            else:
                self.logger.addHandler(handler)

    def _write_buffer(self):
        """
        Write the contents of the buffer to the log.
        """
        line = f"{self._line_prefix}{''.join(self._buffer)}"
        self._buffer.clear()
        super().write(line)

    def write(self, output: Union[bytes, str]):
        """
        Write the given output to the log.
        """
        # Ensure that the output is a string
        if not isinstance(output, str):
            output = str(output, ENCODING, errors="replace")

        for char in output:
            self._buffer.append(char)
            if char == "\n":
                self._write_buffer()

    def cleanup(self):
        """
        Flush the buffer.
        """
        if self._buffer:
            self._buffer.append("\n")
            self._write_buffer()

    @classmethod
    def for_build_container(cls, name: str) -> "ContainerLogger":
        """
        Return a ContainerLogger for a build container.
        """
        color = cls.BUILD_LOG_COLORS.next()
        name_idx = f"{name}_{color}"
        if name_idx not in cls.LOGGERS:
            cls.LOGGERS[name_idx] = ContainerLogger(name_idx, color, prefix=name)
        return cls.LOGGERS[name_idx]

    @classmethod
    def for_service_container(cls, name: str) -> "ContainerLogger":
        """
        Return a ContainerLogger for a service container.
        """
        color = cls.SERVICE_LOG_COLORS.next()
        if name not in cls.LOGGERS:
            cls.LOGGERS[name] = ContainerLogger(f"service-{name}", color)
        return cls.LOGGERS[name]


class DockerPullProgress:
    # Inspired by https://github.com/docker/docker-py/issues/376#issuecomment-1414535176
    tasks = {}

    def __init__(self):
        self.progress = progress.Progress()

    def __enter__(self):
        self.progress.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.progress.__exit__(exc_type, exc_val, exc_tb)

    def status_report(self, data: dict) -> None:
        if not data:
            return
        try:
            status = data.get("status")
            if status == "Downloading":
                color = "cyan"
            elif status == "Extracting":
                color = "green"
            else:
                # Skip other statuses
                return
            status_id = data.get("id")
            if not status_id or status_id == "0":
                return
            description = f"[{color}]{status_id}: {status}"

            if status_id not in self.tasks:
                self.tasks[status_id] = self.progress.add_task(
                    description,
                    total=data["progressDetail"]["total"],
                )
            else:
                self.progress.update(
                    self.tasks[status_id],
                    completed=data["progressDetail"]["current"],
                    description=description,
                )
        except Exception:
            # Do not worry about any exceptions since it's just for helpful display purposes
            pass



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import logging
import os
import platform as python_platform
import random
import re
import shutil
import tempfile
from threading import Thread
import uuid
from queue import SimpleQueue
from typing import Dict, Iterator, List, Optional, Union

import python_on_whales
import timeout_decorator
from python_on_whales import docker
from retry import retry

from buildrunner.config import BuildRunnerConfig
from buildrunner.config.models import MP_LOCAL_REGISTRY
from buildrunner.docker import get_dockerfile
from buildrunner.docker.image_info import BuiltImageInfo, BuiltTaggedImage
from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerError,
    BuildRunnerProcessingError,
)
from buildrunner.docker import new_client


LOGGER = logging.getLogger(__name__)
OUTPUT_LINE = "-----------------------------------------------------------------"
IMAGE_PREFIX = "buildrunner-mp"

PUSH_TIMEOUT = 300


class RegistryInfo:
    """Registry information"""

    def __init__(self, name: str, ip_addr: str, port: int):
        self._name = name
        self._ip_addr = ip_addr
        self._port = port

    @property
    def name(self) -> str:
        """The registry name"""
        return self._name

    @property
    def ip_addr(self) -> str:
        """The registry ip address"""
        return self._ip_addr

    @property
    def port(self) -> int:
        """The registry port"""
        return self._port

    def __str__(self):
        """Returns a string representation of the registry info"""
        return f"{self._name} {self._ip_addr}:{self._port}"

    def __repr__(self):
        """Returns a string representation of the registry info"""
        return self.__str__()


class MultiplatformImageBuilder:  # pylint: disable=too-many-instance-attributes
    """Multiple platform image builder"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        docker_registry: Optional[str] = None,
        build_registry: Optional[str] = MP_LOCAL_REGISTRY,
        temp_dir: str = os.getcwd(),
        platform_builders: Optional[Union[Dict[str, List[str]], Dict[str, str]]] = None,
        cache_builders: Optional[List[str]] = None,
        cache_from: Optional[Union[dict, str]] = None,
        cache_to: Optional[Union[dict, str]] = None,
        secrets: Optional[List[str]] = None,
    ):
        self._docker_registry = docker_registry
        self._build_registry = build_registry
        self._use_local_registry = build_registry == MP_LOCAL_REGISTRY
        self._temp_dir = temp_dir
        self._platform_builders = platform_builders
        self._cache_builders = set(cache_builders if cache_builders else [])
        self._cache_from = cache_from
        self._cache_to = cache_to
        self._secrets = secrets
        if self._cache_from or self._cache_to:
            LOGGER.info(
                f"Configuring multiplatform builds to cache from {cache_from} and to {cache_to} "
                f"for builders {', '.join(cache_builders) if cache_builders else '(all)'}"
            )

        self._built_images: List[BuiltImageInfo] = []
        self._local_registry_is_running = False
        self._mp_registry_info = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._local_registry_is_running:
            self._stop_local_registry()

    def set_cache_from(self, cache_from):
        self._cache_from = cache_from

    def set_cache_to(self, cache_to):
        self._cache_to = cache_to

    def _build_registry_address(self) -> str:
        """Returns the address of the local registry"""
        if self._build_registry == MP_LOCAL_REGISTRY:
            return f"{self._mp_registry_info.ip_addr}:{self._mp_registry_info.port}"
        return self._build_registry

    def _start_local_registry(self):
        """
        Starts a local docker registry container

        Returns:
            str: The name of the registry container
        """
        if not self._local_registry_is_running:
            if os.getenv("BUILDRUNNER_CONTAINER"):
                raise BuildRunnerConfigurationError(
                    "Multiplatform builds cannot be used in the buildrunner Docker image without "
                    "a 'build-registry' configured in the global buildrunner configuration."
                )

            LOGGER.debug("Starting local docker registry")
            image = "registry"
            if self._docker_registry:
                image = f"{self._docker_registry}/{image}"
            container = docker.run(image, detach=True, publish_all=True)
            ports = container.network_settings.ports

            # If any assert fails something changed in the registry image and we need to update this code
            assert len(ports) == 1, f"Expected 1 port, but got {len(ports)}"
            assert isinstance(ports.get("5000/tcp")[0], dict), (
                f"Expected dict, but got {type(ports.get('5000/tcp')[0])}"
            )
            assert ports.get("5000/tcp")[0].get("HostIp") == "0.0.0.0", (
                f"Expected HostIp to be 0.0.0.0 but got {ports.get('5000/tcp')[0].get('HostIp')}"
            )

            self._mp_registry_info = RegistryInfo(
                container.name, "localhost", ports.get("5000/tcp")[0].get("HostPort")
            )
            self._local_registry_is_running = True
            LOGGER.debug(f"Started local registry {self._mp_registry_info}")
        else:
            LOGGER.warning("Local registry is already running")

    def _stop_local_registry(self):
        """
        Stops and removes the local registry along with any images
        """
        if self._local_registry_is_running:
            LOGGER.debug(
                f"Stopping and removing local registry {self._mp_registry_info}"
            )
            try:
                docker.remove(self._mp_registry_info.name, volumes=True, force=True)
            except python_on_whales.exceptions.NoSuchContainer as err:
                LOGGER.error(
                    f"Failed to stop and remove local registry {self._mp_registry_info.name}: {err}"
                )
            self._local_registry_is_running = False
        else:
            LOGGER.warning("Local registry is not running when attempting to stop it")

    def _get_build_cache_options(self, builder: Optional[str]) -> dict:
        cache_options = {
            "cache_from": self._cache_from,
            "cache_to": self._cache_to,
        }
        # If there are no configured cache builders, always return the cache options
        if not self._cache_builders:
            return cache_options

        # If there are cache builders configured, make sure the current builder is in it
        actual_builder = builder or "default"
        return cache_options if actual_builder in self._cache_builders else {}

    def _log_buildx(self, logs_itr: Iterator[str], platform: str):
        for log in logs_itr:
            LOGGER.info(f"[{platform}] {log}".strip())

    # pylint: disable=too-many-arguments,too-many-locals
    def _build_with_inject(
        self,
        inject: dict,
        image_ref: str,
        path: str,
        platform: str,
        dockerfile: str,
        build_args: dict,
        build_kwargs: dict,
        copy_source_path: bool = True,
    ) -> None:
        if not path or not os.path.isdir(path):
            LOGGER.warning(
                f"Failed to inject {inject} for {image_ref} since path {path} isn't a directory."
            )
            return

        dir_prefix = "mp-tmp-dir"
        with tempfile.TemporaryDirectory(
            dir=self._temp_dir, prefix=dir_prefix
        ) as tmp_dir:
            context_dir = os.path.join(tmp_dir, f"{dir_prefix}/")

            if copy_source_path:
                # Copy the entire source path as the build context
                shutil.copytree(
                    path,
                    context_dir,
                    ignore=shutil.ignore_patterns(dir_prefix, ".git"),
                    symlinks=True,
                )
            else:
                # Create an empty context directory
                os.makedirs(context_dir, exist_ok=True)

            for src, dest in inject.items():
                src_path = os.path.join(path, src)

                # Remove '/' prefix
                if dest.startswith("/"):
                    dest = dest[1:]

                dest_path = os.path.join(context_dir, dest)

                # Check to see if the dest dir exists, if not create it
                dest_dir = os.path.dirname(dest_path)
                if not os.path.isdir(dest_dir):
                    os.makedirs(dest_dir, exist_ok=True)

                # Copy source to destination
                if os.path.isdir(src_path):
                    shutil.copytree(src_path, dest_path)
                else:
                    shutil.copy(src_path, dest_path)

            # Dockerfile listed in inject will overwrite the dockerfile passed in
            if dockerfile and "Dockerfile" not in inject.values():
                LOGGER.info(
                    f"Injecting Dockerfile {dockerfile} to context directory '{context_dir}'"
                )
                shutil.copy(dockerfile, f"{context_dir}/Dockerfile")

            assert os.path.isdir(context_dir), (
                f"Failed to create context dir {context_dir}"
            )

            logs_itr = docker.buildx.build(
                context_dir,
                build_args=build_args,
                load=True,
                platforms=[platform],
                stream_logs=True,
                tags=[image_ref],
                **build_kwargs,
                **self._get_build_cache_options(build_kwargs.get("builder")),
            )
            self._log_buildx(logs_itr, platform)

    @staticmethod
    def _get_image_digest(image_ref: str) -> Optional[str]:
        """
        Get the digest for an image with multiple fallback strategies.
        """
        # Strategy 1: Try to get digest from local image first via python-on-whales
        inspect_result = docker.buildx.imagetools.inspect(image_ref)
        if inspect_result and inspect_result.config:
            LOGGER.info(
                f"Found digest for {image_ref} from python-on-whales: {inspect_result.config.digest}"
            )
            return inspect_result.config.digest
        LOGGER.warning(
            f"Failed to inspect image {image_ref}: {inspect_result} using python-on-whales"
        )

        # Strategy 2: Try docker-py APIClient inspect_image (local)
        try:
            docker_client = new_client()
            image_data = docker_client.inspect_image(image_ref)
            # Extract digest from first repo digest that contains "@"
            repo_digests = image_data.get("RepoDigests", [])
            for repo_digest in repo_digests:
                if "@" in repo_digest:
                    digest = repo_digest.split("@", 1)[1]
                    LOGGER.info(
                        f"Found digest for {image_ref} from docker-py inspect_image: {digest}"
                    )
                    return digest
        except Exception as err:
            LOGGER.warning(
                f"Failed to inspect image {image_ref} using docker-py: {err}"
            )

        LOGGER.warning(f"Could not find digest for {image_ref}")
        return None

    # pylint: disable=too-many-arguments
    @retry(
        python_on_whales.exceptions.DockerException,
        tries=5,
        delay=1,
        backoff=3,
        max_delay=30,
        logger=LOGGER,
    )
    def _build_single_image(
        self,
        queue: SimpleQueue,
        image_ref: str,
        platform: str,
        path: str,
        dockerfile: str,
        target: str,
        build_args: dict,
        inject: dict,
        cache: Optional[bool] = None,
        pull: bool = False,
        secrets: Optional[List[str]] = None,
        copy_source_path: bool = True,
    ) -> None:
        """
        Builds a single image for the given platform.

        Args:
            queue (SimpleQueue): The queue to put the message digest on
            image_ref (str): The image ref for the new image
            platform (str): The platform to build the image for (e.g. linux/amd64)
            path (str): The path to the Dockerfile.
            dockerfile (str): The path/name of the Dockerfile (i.e. <path>/Dockerfile).
            target (str): The name of the stage to build in a multi-stage Dockerfile
            build_args (dict): The build args to pass to docker.
            inject (dict): The files to inject into the build context.
            secrets (List[str]): The secrets to pass to docker.
            copy_source_path (bool): Whether to copy the source path into the build context.
        """
        assert os.path.isdir(path) and os.path.exists(dockerfile), (
            f"Either path {path} ({os.path.isdir(path)}) or file "
            f"'{dockerfile}' ({os.path.exists(dockerfile)}) does not exist!"
        )

        # Get the builder for the platform
        builders = (
            self._platform_builders.get(platform) if self._platform_builders else None
        )
        builder = None

        if builders:
            if isinstance(builders, str):
                builder = builders
            elif isinstance(builders, list):
                # Randomly select a builder from the list
                builder = random.choice(builders)
            else:
                raise BuildRunnerConfigurationError(
                    f"Invalid platform builders configuration for platform {platform}"
                )

        LOGGER.debug(f"Building image {image_ref} for platform {platform}")
        LOGGER.info(
            f"Building image for platform {platform} with {builder or 'default'} builder"
        )

        # Build kwargs for the buildx build command
        build_kwargs = {}
        if builder:
            build_kwargs["builder"] = builder
        # Handle both no cache and cache
        if cache is not None:
            build_kwargs["cache"] = cache
        if pull:
            build_kwargs["pull"] = pull
        if secrets:
            build_kwargs["secrets"] = secrets
        if target:
            build_kwargs["target"] = target

        if inject and isinstance(inject, dict):
            self._build_with_inject(
                path=path,
                inject=inject,
                image_ref=image_ref,
                platform=platform,
                dockerfile=dockerfile,
                build_args=build_args,
                build_kwargs=build_kwargs,
                copy_source_path=copy_source_path,
            )
        else:
            logs_itr = docker.buildx.build(
                path,
                tags=[image_ref],
                platforms=[platform],
                load=True,
                file=dockerfile,
                build_args=build_args,
                stream_logs=True,
                **build_kwargs,
                **self._get_build_cache_options(builder),
            )
            self._log_buildx(logs_itr, platform)

        # Push after the initial load to support remote builders that cannot access the local registry
        docker.push([image_ref])

        # Retrieve the digest and put it on the queue
        image_digest = self._get_image_digest(image_ref)
        if image_digest is None:
            raise BuildRunnerProcessingError(
                f"Failed to retrieve digest for image {image_ref} after push. "
                "This may indicate an issue with the registry or image metadata. "
                "Check that the image was pushed successfully and the registry is accessible."
            )
        queue.put((image_ref, image_digest))

    @staticmethod
    def get_native_platform():
        """
        Retrieves the native platform for the current machine or a name that is similar
        to the native platform used by Docker.
        """
        host_system = python_platform.system()
        host_machine = python_platform.machine()

        if host_system.lower() in ("darwin", "linux"):
            host_system = "linux"
        if host_machine.lower() == "x86_64":
            host_machine = "amd64"
        elif host_machine.lower() == "aarch64":
            host_machine = "arm64"
        return f"{host_system}/{host_machine}"

    def _get_single_platform_to_build(self, platforms: List[str]) -> str:
        """Returns the platform to build for single platform flag"""

        assert isinstance(platforms, list), f"Expected list, but got {type(platforms)}"
        assert len(platforms) > 0, (
            f"Expected at least one platform, but got {len(platforms)}"
        )

        native_platform = self.get_native_platform()

        for curr_platform in platforms:
            if native_platform in curr_platform:
                return curr_platform

        return platforms[0]

    # pylint: disable=too-many-locals
    def build_multiple_images(
        self,
        platforms: List[str],
        path: str = ".",
        file: str = "Dockerfile",
        target: Optional[str] = None,
        use_threading: bool = True,
        build_args: dict = None,
        inject: dict = None,
        cache: Optional[bool] = None,
        pull: bool = False,
        secrets: Optional[List[str]] = None,
        copy_source_path: bool = True,
    ) -> BuiltImageInfo:
        """
        Builds multiple images for the given platforms. One image will be built for each platform.

        :arg platforms: The platforms to build the image for (e.g. linux/amd64)
        :arg path: The path to the Dockerfile. Defaults to ".".
        :arg file: The path/name of the Dockerfile (i.e. <path>/Dockerfile). Defaults to "Dockerfile".
        :arg use_threading: Whether to use threading to build the images. Defaults to True.
        :arg build_args: The build args to pass to docker. Defaults to None.
        :arg inject: The files to inject into the build context. Defaults to None.
        :arg cache: If true, enables cache, defaults to False
        :arg pull: If true, pulls image before build, defaults to False
        :arg copy_source_path: Whether to copy the source path into the build context. Defaults to True.
        :return: A BuiltImageInfo instance that describes the built images for each platform and can be used to track
                 final images as well
        """
        if build_args is None:
            build_args = {}
        build_args["DOCKER_REGISTRY"] = self._docker_registry

        # It is not valid to pass None for the path when building multi-platform images
        if not path:
            if os.path.exists(file):
                path = os.path.dirname(file)
            else:
                path = "."

        dockerfile, cleanup_dockerfile = get_dockerfile(file)

        # Track this newly built image
        built_image = BuiltImageInfo(id=str(uuid.uuid4()))
        self._built_images.append(built_image)

        LOGGER.debug(
            f"Building multi-platform images {built_image} for platforms {platforms} from {dockerfile}"
        )

        if self._use_local_registry and not self._local_registry_is_running:
            # Starts local registry container to do ephemeral image storage
            self._start_local_registry()

        # Uses the current node name as the repo
        if python_platform.node():
            sanitized_name = f"{IMAGE_PREFIX}-{re.sub(r'[^a-zA-Z0-9]', '', python_platform.node()).lower()}"
        else:
            sanitized_name = f"{IMAGE_PREFIX}-unknown-node"
        repo = f"{self._build_registry_address()}/{sanitized_name}"

        if BuildRunnerConfig.get_instance().global_config.disable_multi_platform:
            platforms = [self._get_single_platform_to_build(platforms)]
            LOGGER.info(OUTPUT_LINE)
            LOGGER.info(
                "Note: Disabling multi-platform build, this will only build a single-platform image."
            )
            LOGGER.info(f"image: {repo} platform:{platforms[0]}")
            LOGGER.info(OUTPUT_LINE)
        else:
            LOGGER.info(OUTPUT_LINE)
            LOGGER.info(
                "Note: Building multi-platform images can take a long time, please be patient."
            )
            LOGGER.info(
                "If you are running this locally, you can speed this up by using the '--disable-multi-platform' "
                "CLI flag or set the 'disable-multi-platform' flag in the global config file."
            )
            LOGGER.info(OUTPUT_LINE)

        threads = []
        LOGGER.info(
            f"Starting builds for {len(platforms)} platforms in {'parallel' if use_threading else 'sequence'}"
        )

        # Since threading may be used, use a simple queue to communicate with the build method
        # This queue receives tuples of (image_ref, image_digest)
        image_info_by_image_ref = {}
        queue = SimpleQueue()
        for platform in platforms:
            tag = f"{built_image.id}-{platform.replace('/', '-')}"
            image_ref = f"{repo}:{tag}"
            # Contains all fields needed for the BuiltTaggedImage instance except digest, which will be added later
            image_info_by_image_ref[image_ref] = {
                "repo": repo,
                "tag": tag,
                "platform": platform,
            }
            build_single_image_args = (
                queue,
                image_ref,
                platform,
                path,
                dockerfile,
                target,
                build_args,
                inject,
                cache,
                pull,
                secrets,
                copy_source_path,
            )
            LOGGER.debug(f"Building {repo} for {platform}")
            if use_threading:
                threads.append(
                    Thread(
                        target=self._build_single_image,
                        args=build_single_image_args,
                    )
                )
            else:
                self._build_single_image(*build_single_image_args)

        # Start and join threads in parallel if enabled
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

        while not queue.empty():
            image_ref, image_digest = queue.get()
            assert image_ref in image_info_by_image_ref, (
                f"Image ref {image_ref} is missing in generated info"
            )
            image_info = image_info_by_image_ref.pop(image_ref)
            built_image.add_platform_image(
                image_info.get("platform"),
                BuiltTaggedImage(
                    **image_info,
                    digest=image_digest,
                ),
            )
        assert not image_info_by_image_ref, (
            f"Image refs were not generated successfully, unclaimed refs: {image_info_by_image_ref}"
        )

        if cleanup_dockerfile and dockerfile and os.path.exists(dockerfile):
            os.remove(dockerfile)

        return built_image

    @retry(
        TimeoutError,
        tries=5,
        logger=LOGGER,
    )
    @timeout_decorator.timeout(PUSH_TIMEOUT)
    def _push_with_timeout_and_retries(
        self, src_names: List[str], tag_names: List[str]
    ) -> None:
        """
        Creates tags from a set of source images in the remote registry.
        This method will time out if it takes too long. An exception may be
        caught and retried for the timeout.

        Args:
            src_names (List[str]): The source images to combine into the image manifest
            tag_names (List[str]): The tags to push with the final image manifest

        Raises:
            TimeoutError: If the image fails to push within the timeout
        """
        LOGGER.info(f"Pushing sources {src_names} to tags {tag_names}")
        docker.buildx.imagetools.create(sources=src_names, tags=tag_names)

    def push(self) -> None:
        """
        Pushes all built images to their tagged image counterparts.
        :raises TimeoutError: If the image fails to push within the specified timeout and retries
        """
        push_count = 0
        expected_pushes = 0
        for built_image in self._built_images:
            for tagged_image in built_image.tagged_images:
                expected_pushes += len(tagged_image.image_refs)

        for built_image in self._built_images:
            if not built_image.tagged_images:
                LOGGER.info(
                    f"No tags exist for built image {built_image.id}, not pushing"
                )
                continue

            source_image_refs = [image.image_ref for image in built_image.built_images]

            for tagged_image in built_image.tagged_images:
                LOGGER.info(
                    f"Pushing {built_image.id} to {', '.join(tagged_image.image_refs)}"
                )

                LOGGER.debug(f"Creating manifest(s) {tagged_image}")
                # Push each tag individually in order to prevent strange errors with multiple matching tags
                for image_ref in tagged_image.image_refs:
                    self._push_with_timeout_and_retries(source_image_refs, [image_ref])
                    push_count += 1
                LOGGER.info(
                    f"Successfully pushed multiplatform image(s) {tagged_image}"
                )

        if push_count != expected_pushes:
            raise BuildRunnerError(
                f"Failed to push all images. Expected to push {expected_pushes} but pushed {push_count}."
            )

    @property
    def num_built_images(self) -> int:
        """
        Returns the count of built images that currently exist.
        """
        return len(self._built_images)

    @staticmethod
    def tag_native_platform(built_image: BuiltImageInfo) -> None:
        """
        Tags a single built image into the host registry for the platform matching this machine's platform.
        If a matching image to the native platform is not found, then the first image is tagged. The tags come
        from all registered tagged images in the built image info.
        :arg built_image: The built image to pull and tag
        """
        for tagged_image in built_image.tagged_images:
            LOGGER.debug(
                f"Tagging {built_image} as {tagged_image} for the native platform"
            )
            native_image = built_image.native_platform_image

            # Pull the native image and then tag it
            try:
                docker.pull(native_image.image_ref)
                for tagged_image_ref in tagged_image.image_refs:
                    docker.tag(native_image.image_ref, tagged_image_ref)
                    LOGGER.info(
                        f"Tagged {native_image.image_ref} as {tagged_image_ref}"
                    )
                docker.image.remove(native_image.image_ref, force=True)
            except python_on_whales.exceptions.DockerException as err:
                LOGGER.error(
                    f"Failed while tagging {native_image.image_ref} as {tagged_image}: {err}"
                )
                raise err



"""
Copyright 2024 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import glob
import os
import re


from buildrunner import loggers
import buildrunner.docker
from buildrunner.config import BuildRunnerConfig
from buildrunner.config.models_step import StepBuild
from buildrunner.docker.importer import DockerImporter
import buildrunner.docker.builder
from buildrunner.errors import (
    BuildRunnerConfigurationError,
)
from buildrunner.steprunner.tasks import BuildStepRunnerTask

LOGGER = loggers.ConsoleLogger(__name__)


class BuildBuildStepRunnerTask(BuildStepRunnerTask):  # pylint: disable=too-many-instance-attributes
    """
    Class used to manage "build" build tasks.
    """

    DOCKERFILE_IMAGE_REGEX = re.compile(r"^FROM (.*)", re.IGNORECASE)

    def __init__(
        self,
        step_runner,
        step: StepBuild,
        image_to_prepend_to_dockerfile=None,
        force_legacy_builder=False,
    ):  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
        super().__init__(step_runner, step)
        self._docker_client = buildrunner.docker.new_client(
            timeout=step_runner.build_runner.docker_timeout,
        )
        self.to_inject = {}
        self.image_to_prepend_to_dockerfile = image_to_prepend_to_dockerfile
        self.force_legacy_builder = force_legacy_builder

        self._import = step.import_param
        self.path = step.path
        # Track whether path was explicitly specified in the config
        self._path_explicitly_specified = step.path is not None
        self.dockerfile = step.dockerfile
        self.target = step.target
        self.nocache = step.no_cache
        self.platform = step.platform
        self.platforms = step.platforms
        self.buildargs = step.buildargs
        self.cache_from = step.cache_from

        buildrunner_config = BuildRunnerConfig.get_instance()

        #  Pull the cache_from images only for single platform builds
        if self.cache_from and isinstance(self.cache_from, list) and not self.platforms:
            for cache_from_image in self.cache_from:
                if isinstance(cache_from_image, str):
                    try:
                        self._docker_client.pull(
                            cache_from_image, platform=self.platform
                        )
                        # If the pull is successful, add the image to be cleaned up at the end of the script
                        self.step_runner.build_runner.generated_images.append(
                            cache_from_image
                        )
                        self.step_runner.log.write(
                            f"Using cache_from image: {cache_from_image}\n"
                        )
                    except Exception:  # pylint: disable=broad-except
                        self.step_runner.log.write(
                            f"WARNING: Unable to pull the cache_from image: {cache_from_image}\n"
                        )

        for src_glob, dest_path in (step.inject or {}).items():
            if not dest_path:
                dest_path = ""
            _src_glob = buildrunner_config.to_abs_path(src_glob)
            xsglob = glob.glob(_src_glob)
            if not xsglob:
                # Failed to resolve the glob
                raise BuildRunnerConfigurationError(
                    f"Unable to expand inject glob: {_src_glob}"
                )
            if len(xsglob) == 1:
                # Only one source - destination may be directory or filename - check for a trailing
                # '/' and treat it accordingly.
                source_file = xsglob[0]
                if dest_path and (
                    dest_path[-1] == "/" or os.path.split(dest_path)[-1] in (".", "..")
                ):
                    self.to_inject[source_file] = os.path.normpath(
                        os.path.join(".", dest_path, os.path.basename(source_file))
                    )
                else:
                    self.to_inject[source_file] = os.path.normpath(
                        os.path.join(".", dest_path)
                    )
            else:
                # Multiple sources - destination *must* be a directory - add the source basename
                # to the dest_dir name.
                for source_file in xsglob:
                    self.to_inject[source_file] = os.path.normpath(
                        os.path.join(
                            ".",
                            dest_path,
                            os.path.basename(source_file),
                        )
                    )

        if not self._import and not any((self.path, self.dockerfile, self.to_inject)):
            raise BuildRunnerConfigurationError(
                'Docker build context must specify a "path", "dockerfile", or "inject" attribute'
            )

        if self.path:
            self.path = buildrunner_config.to_abs_path(self.path)

        if self.path and not os.path.exists(self.path):
            raise BuildRunnerConfigurationError(
                f"Step {self.step_runner}:build:path:{self.path}: Invalid build context path"
            )

        if not self.dockerfile:
            if self.path:
                path_dockerfile = os.path.join(self.path, "Dockerfile")
                if os.path.exists(path_dockerfile):
                    self.dockerfile = path_dockerfile
            for src_file, dest_file in self.to_inject.items():
                if os.path.normpath(dest_file) in [
                    "Dockerfile",
                    "/Dockerfile",
                ]:
                    self.dockerfile = src_file

        dockerfile_image = None
        if self.dockerfile:
            _dockerfile_abs_path = buildrunner_config.to_abs_path(
                self.dockerfile,
            )
            if os.path.exists(_dockerfile_abs_path):
                self.dockerfile = _dockerfile_abs_path
                if self.image_to_prepend_to_dockerfile:
                    # need to load the contents of the Dockerfile so that we
                    # can prepend the image
                    with open(
                        _dockerfile_abs_path, "r", encoding="utf-8"
                    ) as _dockerfile:
                        self.dockerfile = _dockerfile.read()

            if self.image_to_prepend_to_dockerfile:
                # prepend the given image to the dockerfile
                self.dockerfile = (
                    f"FROM {self.image_to_prepend_to_dockerfile}\n{self.dockerfile}"
                )
                dockerfile_image = self.image_to_prepend_to_dockerfile
            else:
                # Find from image in dockerfile
                match = self.DOCKERFILE_IMAGE_REGEX.match(self.dockerfile)
                if match:
                    dockerfile_image = match.group(1)

        # Set the pull attribute based on configuration or the image itself
        if step.pull is not None:
            self.pull = step.pull
            self.step_runner.log.write(
                f"Pulling image was overridden via config to {self.pull}\n"
            )
        elif not dockerfile_image:
            # Default to true if we could not determine the image
            self.pull = True
            self.step_runner.log.write(
                "Could not determine docker image, defaulting pull to true\n"
            )
        else:
            # If the image was previously committed in this run, do not pull by default
            self.pull = (
                dockerfile_image not in self.step_runner.build_runner.committed_images
            )
            self.step_runner.log.write(
                f"Pull was not specified in configuration, defaulting to {self.pull}\n"
            )

    def run(self, context):
        # 'import' will override other configuration and perform a 'docker
        # import'
        if self._import and not self.platforms:
            self.step_runner.log.write(
                f"  Importing {self._import} as a Docker image\n"
            )
            context["image"] = DockerImporter(
                self._import,
                timeout=self.step_runner.build_runner.docker_timeout,
            ).import_image()
            return

        if not self.dockerfile:
            raise BuildRunnerConfigurationError(
                "Cannot find a Dockerfile in the given path or inject configurations"
            )

        buildrunner_config = BuildRunnerConfig.get_instance()
        docker_registry = buildrunner_config.global_config.docker_registry
        self.step_runner.log.write("Running docker build\n")

        try:
            if (
                self.platforms or not buildrunner_config.run_config.use_legacy_builder
            ) and not self.force_legacy_builder:
                if self.platforms and buildrunner_config.run_config.use_legacy_builder:
                    LOGGER.warning(
                        f"Ignoring use-legacy-builder. Using the legacy builder for multiplatform images {self.platforms} is not supported. "
                        "If you are want to use the legacy builder and specify the platform, please use 'platform' in the configuration file."
                    )

                # Setting the platform if it is set in the step
                if self.platform:
                    LOGGER.info(f"Setting platform to {self.platform}")
                    self.platforms = [self.platform]

                if not self.platforms:
                    native_platform = (
                        self.step_runner.multi_platform.get_native_platform()
                    )
                    LOGGER.info(f"Setting platforms to [{native_platform}]")
                    self.platforms = [native_platform]

                # Build the image
                built_images = self.step_runner.multi_platform.build_multiple_images(
                    platforms=self.platforms,
                    path=self.path,
                    file=self.dockerfile,
                    target=self.target,
                    build_args=self.buildargs,
                    inject=self.to_inject,
                    cache=not self.nocache,
                    pull=self.pull,
                    secrets=self.step.secrets,
                    copy_source_path=self._path_explicitly_specified,
                )

                # Set expected number of platforms
                expected_num_platforms = 1
                if (
                    self.platforms
                    and not buildrunner_config.global_config.disable_multi_platform
                ):
                    expected_num_platforms = len(self.platforms)

                # Set the number of built platforms
                num_built_platforms = len(built_images.platforms)

                # Compare the number of built platforms with the expected number of platforms
                assert num_built_platforms == expected_num_platforms, (
                    f"Number of built images ({num_built_platforms}) does not match "
                    f"the number of expected platforms ({expected_num_platforms})"
                )
                context["mp_built_image"] = built_images
                if num_built_platforms > 0:
                    context["image"] = built_images.native_platform_image.image_ref

            else:
                # Use the legacy builder
                image = buildrunner.docker.builder.build_image(
                    path=self.path,
                    inject=self.to_inject,
                    dockerfile=self.dockerfile,
                    docker_registry=docker_registry,
                    console=self.step_runner.log,
                    nocache=self.nocache,
                    cache_from=self.cache_from,
                    pull=self.pull,
                    buildargs=self.buildargs,
                    platform=self.platform,
                    target=self.target,
                    copy_source_path=self._path_explicitly_specified,
                )
                context["image"] = image
                self.step_runner.build_runner.generated_images.append(image)
        except Exception as exc:
            self.step_runner.log.write(f"ERROR: {exc}\n")
            raise



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
from io import StringIO

from fabric import Connection

from buildrunner.errors import (
    BuildRunnerProcessingError,
)
from buildrunner.steprunner.tasks import BuildStepRunnerTask
from buildrunner.config import BuildRunnerConfig
from buildrunner.config.models_step import StepRemote


class RemoteBuildStepRunnerTask(BuildStepRunnerTask):
    """
    Class used to manage "remote" build tasks.
    """

    def __init__(self, step_runner, step: StepRemote):
        super().__init__(step_runner, step)

        self.host = BuildRunnerConfig.get_instance().get_build_server_from_alias(
            step.host,
        )
        self.cmd = step.cmd
        self.artifacts = step.artifacts

    def run(self, context):  # pylint: disable=unused-argument,too-many-branches,too-many-locals
        self.step_runner.log.write(f"Building on remote host {self.host}\n\n")

        # call remote functions to copy tar and build remotely
        remote_build_dir = f"/tmp/buildrunner/{self.step_runner.build_runner.build_id}-{self.step_runner.name}"
        remote_archive_filepath = remote_build_dir + "/source.tar"
        self.step_runner.log.write(
            f"[{self.host}] Creating temporary remote directory '{remote_build_dir}'\n"
        )

        with Connection(self.host) as connection:
            mkdir_result = connection.run(
                f"mkdir -p {remote_build_dir}",
                warn=True,
                out_stream=self.step_runner.log,
                err_stream=self.step_runner.log,
            )
            if mkdir_result.return_code:
                raise BuildRunnerProcessingError("Error creating remote directory")

            try:  # pylint: disable=too-many-nested-blocks
                self.step_runner.log.write(
                    f"[{self.host}] Pushing archive file to remote directory\n"
                )
                files = connection.put(
                    self.step_runner.build_runner.get_source_archive_path(),
                    remote_archive_filepath,
                )
                if files:
                    self.step_runner.log.write(
                        f"[{self.host}] Extracting source tree archive on remote host:\n"
                    )
                    extract_result = connection.run(
                        f"(cd {remote_build_dir}; tar -xvf source.tar && rm -f source.tar)",
                        warn=True,
                        out_stream=self.step_runner.log,
                        err_stream=self.step_runner.log,
                    )
                    if extract_result.return_code:
                        raise BuildRunnerProcessingError(
                            "Error extracting archive file"
                        )

                    self.step_runner.log.write(
                        f"[{self.host}] Running command '{self.cmd}'\n"
                    )
                    package_result = connection.run(
                        f"(cd {remote_build_dir}; {self.cmd})",
                        warn=True,
                        out_stream=self.step_runner.log,
                        err_stream=self.step_runner.log,
                    )

                    if self.artifacts:
                        _arts = []
                        for _art, _props in self.artifacts.items():
                            # check to see if there are artifacts
                            # that match the pattern
                            dummy_out = StringIO()
                            art_result = connection.run(
                                f"ls -A1 {remote_build_dir}/{_art}",
                                hide=True,
                                warn=True,
                                out_stream=dummy_out,
                                err_stream=dummy_out,
                            )
                            if art_result.return_code:
                                continue

                            # we have at least one match--run the get
                            for _ca in connection.get(  # pylint: disable=not-an-iterable
                                f"{remote_build_dir}/{_art}",
                                f"{self.step_runner.results_dir}/%(basename)s",
                            ):
                                _arts.append(_ca)
                                self.step_runner.build_runner.add_artifact(
                                    os.path.join(
                                        self.step_runner.name,
                                        os.path.basename(_ca),
                                    ),
                                    _props,
                                )
                        self.step_runner.log.write("\nGathered artifacts:\n")
                        for _art in _arts:
                            self.step_runner.log.debug(
                                f"- found {os.path.basename(_art)}",
                            )
                        self.step_runner.log.write("\n")

                    if package_result.return_code:
                        raise BuildRunnerProcessingError("Error running remote build")

                else:
                    raise BuildRunnerProcessingError(
                        "Error uploading source archive to host"
                    )
            finally:
                self.step_runner.log.write(
                    f"[{self.host}] Cleaning up remote temp directory {remote_build_dir}\n"
                )
                cleanup_result = connection.run(
                    f"rm -Rf {remote_build_dir}",
                    out_stream=self.step_runner.log,
                    err_stream=self.step_runner.log,
                )
                if cleanup_result.return_code:
                    raise BuildRunnerProcessingError(
                        "Error cleaning up remote directory"
                    )



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from buildrunner.errors import BuildRunnerProvisionerError


class ShellProvisioner:
    """
    Provisioner used to run a shell script.
    """

    def __init__(self, script, console=None):
        self.script = script
        self.console = console

    def provision(self, runner):
        """
        Run the shell script as a provisioner, with the given shell.
        """
        if self.console:
            self.console.write("Running shell provisioner...\n")
        exit_code = runner.run_script(
            self.script,
            console=self.console,
        )
        if exit_code != 0:
            raise BuildRunnerProvisionerError("Shell provisioner failed")



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import urllib.parse
from typing import Optional

from ..models import GlobalConfig
from . import github
from . import http
from . import file


def fetch_file(url: str, config: Optional[GlobalConfig]) -> str:
    """
    Fetch a file from a provider.
    """

    # pylint: disable=fixme
    # FIXME: the handled checking should be in each handler module (possibly handle_file(parsed_url,
    # config) => bool)

    parsed_url = urllib.parse.urlparse(url)
    scheme = parsed_url.scheme
    if isinstance(scheme, bytes):
        scheme = scheme.decode("utf8")

    if scheme == "github":
        file_contents = github.fetch_file(parsed_url, config)

    elif scheme == "file" or (
        scheme == "" and (parsed_url.netloc == "" or parsed_url.netloc == b"")
    ):
        purl = list(parsed_url)
        purl[0] = "file"
        parsed_url = urllib.parse.ParseResult(*purl)
        file_contents = file.fetch_file(parsed_url, config)

    elif scheme in ("http", "https"):
        file_contents = http.fetch_file(parsed_url, config)

    else:
        raise NotImplementedError(f"Unknown fetch backend: {scheme}")

    return file_contents


# Local Variables:
# fill-column: 100
# End:
